import sys
from acqgenlib import parse_blocks, names, report

def get_int_property(prop):
    """
    Try to derive an integer value from the given property. Return an error if
    such a value cannot be produced, else return the value.
    """
    if prop.type != 'intnumber':
        msg = "Integer value expected for parameter '{}'".format(prop.name)
        report.report_error([prop.lineno], msg)
    return prop.value

def get_nonneg_int_property(prop, exclude_zero=False):
    """
    Try to derive an integer value from the given property. Return an error if
    such a value cannot be produced or if it is negative, else return the value.
    """
    val = get_int_property(prop)
    if exclude_zero:
        if val <= 0:
            msg = "Non-zero positive value expected for parameter '{}'".format(prop.name)
            report.report_error([prop.lineno], msg)
    else:
        if val < 0:
            msg = "Non-negative value expected for parameter '{}'".format(prop.name)
            report.report_error([prop.lineno], msg)

    return val

def get_string_property(prop):
    """
    Try to derive a string value from the given property. Return an error if
    such a value cannot be produced, else return the value.
    """
    if prop.type != 'string':
        msg = "String value expected for parameter '{}'".format(prop.name)
        report.report_error([prop.lineno], msg)
    return prop.value

def get_real_property(prop):
    """
    Try to derive an real value from the given property. Return an error if
    such a value cannot be produced, else return the value.
    """
    if prop.type == 'realnumber':
        return prop.value

    if prop.type == 'intnumber':
        return float(prop.value)

    msg = "Real value expected for parameter '{}'".format(prop.name)
    report.report_error([prop.lineno], msg)

def get_nonneg_real_property(prop, exclude_zero=False):
    """
    Try to derive a real value from the given property. Return an error if
    such a value cannot be produced or if it is negative, else return the value.
    """
    val = get_real_property(prop)
    if exclude_zero:
        if val <= 0:
            msg = "Non-zero positive value expected for parameter '{}'".format(prop.name)
            report.report_error([prop.lineno], msg)
    else:
        if val < 0:
            msg = "Non-negative value expected for parameter '{}'".format(prop.name)
            report.report_error([prop.lineno], msg)

    return val

def get_reallist_property(prop):
    """
    Try to derive a list with real values from the given property. Return an error if
    such a value cannot be produced, else return the value.
    """
    if prop.type != 'reallist':
        msg = "List with real numbers expected for parameter '{}'".format(prop.name)
        report.report_error([prop.lineno], msg)
    return prop.value

def get_identifier_property(prop):
    """
    Try to derive a non-signal name from the given property. Return an error if
    such a value cannot be produced, else return the value.
    """
    if prop.type == 'name':
        value = prop.value
        if value.dev_name is None and value.dev_num is None and value.iop_num is None:
            return value.iop_name

    msg = "Identifier expected for parameter '{}'".format(prop.name)
    report.report_error([prop.lineno], msg)

def get_signal_property(prop):
    """
    Try to derive a signal name from the given property. Return an error if
    such a value cannot be produced, else return the value.
    """
    if prop.type != 'name':
        msg = "Signal expected for parameter '{}'".format(prop.name)
        report.report_error([prop.lineno], msg)

    return prop.value

def get_bool_property(prop):
    """
    Try to derive a boolean value from the given property. Return an error if
    such a value cannot be produced, else return the value.
    """
    if prop.type == 'intnumber':
        return prop.value != 0
    if prop.type == 'name':
        name = prop.value.iop_name
        lname = name.lower()
        if lname in ('true', 'yes', 'on'):
            return True
        if lname in ('false', 'no', 'off'):
            return False

    msg = "Boolean value expected for parameter '{}".format(prop.name)
    report.report_error([prop.lineno], msg)


# {{{ class Configuration:
class Configuration:
    """
    Block for specifying global configuration data.

    @ivar memsize: Amount of available memory (number of bytes).
    @ivar memsize: C{int}

    @ivar eth_port: Name of the ethernet port device.
    @type eth_port: C{str} (C{None} until initialized)

    @ivar delay: Delay length between sampling of the hardware devices, in nano-seconds.
    @type delay: C{int}

    @ivar template: Name of the template file.
    @type template: C{str} or C{None}

    @ivar buffered: Whether to generate a buffered data acquisition program.
    @type buffered: C{bool}

    @ivar default_bufsize: Default buffer size for both from_matlab and to_matlab blocks, if provided.
    @type default_bufsize: C{None} or C{int}

    @ivar default_to_bufsize: Default buffer size for to_matlab blocks, if provided.
    @type default_to_bufsize: C{None} or C{int}

    @ivar default_prewrite: Default prewrite size for to_matlab blocks, if provided.
    @type default_prewrite: C{None} or C{real}

    @ivar default_from_bufsize: Default buffer size for from_matlab blocks, if provided.
    @type default_from_bufsize: C{None} or C{int}

    @ivar default_bufcount: Default number of buffers for both from_matlab and to_matlab blocks, if provided.
    @type default_bufcount: C{None} or C{int}

    @ivar default_to_bufcount: Default number of buffers for to_matlab blocks, if provided.
    @type default_to_bufcount: C{None} or C{int}

    @ivar default_from_bufcount: Default number of buffers for from_matlab blocks, if provided.
    @type default_from_bufcount: C{None} or C{int}

    @ivar default_to_synchronous: Default value of the 'synchronous' property
                                     for to_matlab blocks.
    @type default_to_synchronous: C{bool}

    @ivar matlab_base: Path to the installed Matlab, if specified.
    @ivar matlab_mex: Path to the 'mex' program, if specified.
    @ivar soem_base: Path to the installed SOEM, if specified.
    @ivar soem_lib: Path to the libsoem.a file, if specified.
    @ivar soem_lflags: Linker flags for linking the SOEM library, if specified
    @ivar soem_include: Include directory of the SOEM library, if specified.
    @ivar compiler: C compiler to use, if specified.
    @ivar acqgen_dir: Directory with runtime acqgen source code, if specified.
    @ivar acqgen_sources: Files to use for the runtime acqgen source code, if specified.
    @ivar cflags: Compile flags of the compiler, if specified.
    @ivar lflags: Link flags of the compiler, if specified.
    """
    def __init__(self, lineno, properties):
        self.lineno = lineno

        self.memsize = 256*1024 # Assume plenty of memory
        self.eth_port = None # Must be defined.
        self.delay = 10000000 # 100Hz
        self.template = None # May be overridden in the input file.
        self.buffered = True # Run in buffered mode.
        self.default_bufsize = None
        self.default_to_bufsize = None
        self.default_from_bufsize = None
        self.default_bufcount = None
        self.default_to_bufcount = None
        self.default_from_bufcount = None
        self.default_prewrite = None
        self.default_to_synchronous = False # Synchronous input doesn't make too much sense in general.
        self.matlab_base = None
        self.matlab_mex = None
        self.soem_base = None
        self.soem_lib = None
        self.soem_lflags = None
        self.soem_include = None
        self.compiler = None
        self.acqgen_dir = None
        self.acqgen_sources = None
        self.cflags = None
        self.lflags = None

        for prop in properties:
            if prop.name == 'memsize':
                self.memsize = get_nonneg_int_property(prop, True)
            elif prop.name == 'nic':
                self.eth_port = get_string_property(prop)
            elif prop.name == 'frequency':
                self.delay = int(1000000000 / get_nonneg_real_property(prop, True))
            elif prop.name == 'delay_nano':
                self.delay = get_nonneg_int_property(prop, True)
            elif prop.name == 'delay_micro':
                self.delay = get_nonneg_int_property(prop, True) * 1000
            elif prop.name == 'delay_mili':
                self.delay = get_nonneg_int_property(prop, True) * 1000000
            elif prop.name == 'template':
                self.template = get_string_property(prop)
            elif prop.name == 'buffered':
                self.buffered = get_bool_property(prop)
            elif prop.name == 'default_buffersize':
                self.default_bufsize = get_nonneg_int_property(prop)
            elif prop.name == 'default_to_buffersize':
                self.default_to_bufsize = get_nonneg_int_property(prop)
            elif prop.name == 'default_from_buffersize':
                self.default_from_bufsize = get_nonneg_int_property(prop)
            elif prop.name == 'default_buffercount':
                self.default_bufcount = get_nonneg_int_property(prop)
            elif prop.name == 'default_to_buffercount':
                self.default_to_bufcount = get_nonneg_int_property(prop)
            elif prop.name == 'default_from_buffercount':
                self.default_from_bufcount = get_nonneg_int_property(prop)
            elif prop.name == 'default_prewrite':
                self.default_prewrite = get_nonneg_real_property(prop)
            elif prop.name == 'default_to_synchronous':
                self.default_to_synchronous = get_bool_property(prop)
            elif prop.name == 'matlab_base':
                self.matlab_base = get_string_property(prop)
            elif prop.name == 'matlab_mex':
                self.matlab_mex = get_string_property(prop)
            elif prop.name == 'soem_base':
                self.soem_base = get_string_property(prop)
            elif prop.name == 'soem_lib':
                self.soem_lib = get_string_property(prop)
            elif prop.name == 'soem_lflags':
                self.soem_lflags = get_string_property(prop)
            elif prop.name == 'soem_include':
                self.soem_include = get_string_property(prop)
            elif prop.name == 'compiler':
                self.compiler = get_string_property(prop)
            elif prop.name == 'acqgen_dir':
                self.acqgen_dir = get_string_property(prop)
            elif prop.name == 'acqgen_sources':
                self.acqgen_sources = get_string_property(prop)
            elif prop.name == 'cflags':
                self.cflags = get_string_property(prop)
            elif prop.name == 'lflags':
                self.lflags = get_string_property(prop)
            else:
                msg = "Unknown property '{}' of configuration block".format(prop.name)
                report.report_error([self.lineno], msg)

        if self.eth_port is None:
            msg = "Missing 'nic' (network port name) property of configuration block"
            report.report_error([self.lineno], msg)

        if self.delay > 50000000: # 20Hz
            msg = "Sample rate is low, you may get data sync errors"
            report.report_error([self.lineno], msg, type="warning")

    def get_name(self):
        return 'configuration'

    def get_in_values(self):
        return []

    def get_out_values(self):
        return []
# }}}
# {{{ class SoemInputs:
class SoemInputs:
    """
    Block for reading input variables from Soem master.

    @ivar lineno: Line number of the block in the source file.
    @ivar out_values: Output values.
    """
    def __init__(self, lineno, out_values):
        self.lineno = lineno
        self.out_values = out_values

    def get_name(self):
        return 'soem-input'

    def get_in_values(self):
        return []

    def get_out_values(self):
        return self.out_values

    def set_inout_values(self, in_values, out_values):
        assert len(in_values) == 0
        self.out_values = out_values
# }}}
# {{{ class SoemOutputs:
class SoemOutputs:
    """
    Block for writing output variables to Soem master.

    @ivar lineno: Line number of the block in the source file.
    @ivar in_values: Input values.
    """
    def __init__(self, lineno, in_values):
        self.lineno = lineno
        self.in_values = in_values

    def get_name(self):
        return 'soem-output'

    def get_in_values(self):
        return self.in_values

    def get_out_values(self):
        return []

    def set_inout_values(self, in_values, out_values):
        assert len(out_values) == 0
        self.in_values = in_values
# }}}
# {{{ class FromMatlabBlock:
class FromMatlabBlock:
    """
    Signal group supplying values to the experiment.

    @ivar lineno: Line number of the block in the source file.
    @ivar lineno: C{int}

    @ivar out_values: Output values.
    @type out_values: C{list} of L{Name}

    @ivar index: User index number of the signal group. May be C{None} if no index is provided.
    @ivar buffersize: Size of a buffer for the signal group if specified.
    @ivar buffercount: Number of buffers for the signal group if specified.
    @ivar ring_index: Number of the ring (unique, non-negative, ascending).
    """
    def __init__(self, lineno, out_values, properties, ring_index):
        self.lineno = lineno
        self.out_values = out_values
        self.ring_index = ring_index

        self.index = None
        self.buffersize = None
        self.buffercount = None
        for prop in properties:
            if prop.name == 'index':
                self.index = get_nonneg_int_property(prop, True)
            elif prop.name == 'buffersize':
                self.buffersize = get_nonneg_int_property(prop)
            elif prop.name == 'buffercount':
                self.buffercount = get_nonneg_int_property(prop)
            else:
                msg = "Unknown property '{}' of from_matlab block".format(prop.name)
                report.report_error([self.lineno], msg)

    def get_name(self):
        return 'from-matlab'

    def get_in_values(self):
        return []

    def get_out_values(self):
        return self.out_values

    def set_inout_values(self, in_values, out_values):
        assert len(in_values) == 0
        self.out_values = out_values
# }}}
# {{{ class ToMatlabBlock:
class ToMatlabBlock:
    """
    Signal group taking measurements from the experiment.

    @ivar lineno: Line number of the block in the source file.
    @type lineno: C{int}

    @ivar in_values: Input values.
    @type in_values: C{list} of L{Name}

    @ivar index: User index number of the signal group. May be C{None} if no index is provided.
    @ivar buffersize: Size of a buffer for the signal group.
    @ivar buffercount: Number of buffers for the signal group if specified.
    @ivar prewrite: Fraction of data prewritten to the buffer if specified.
    @ivar enable: Signal controlling enabledness of the group if specified.
    @ivar synchronous: Require matlab getting a buffer with each call.
    @ivar ring_index: Number of the ring (unique, non-negative, ascending).
    """
    def __init__(self, lineno, in_values, properties, ring_index):
        self.lineno = lineno
        self.in_values = in_values
        self.ring_index = ring_index

        self.index = None
        self.buffersize = None
        self.buffercount = None
        self.prewrite = None
        self.enable = None
        self.synchronous = None
        for prop in properties:
            if prop.name == 'index':
                self.index = get_nonneg_int_property(prop, True)
            elif prop.name == 'buffersize':
                self.buffersize = get_nonneg_int_property(prop)
            elif prop.name == 'buffercount':
                self.buffercount = get_nonneg_int_property(prop)
            elif prop.name == 'enable':
                self.enable = get_signal_property(prop)
            elif prop.name == 'prewrite':
                self.prewrite = get_nonneg_real_property(prop)
            elif prop.name == 'synchronous':
                self.synchronous = get_bool_property(prop)
            else:
                msg = "Unknown property '{}' of to_matlab block".format(prop.name)
                report.report_error([prop.lineno], msg)

    def get_name(self):
        return 'to-matlab'

    def get_in_values(self):
        if self.enable is None:
            return self.in_values
        else:
            return self.in_values + [self.enable]

    def get_out_values(self):
        return []

    def set_inout_values(self, in_values, out_values):
        assert len(out_values) == 0

        if self.enable is None:
            self.in_values = in_values
        else:
            assert self.enable in in_values
            self.in_values = [val for val in in_values if val != self.enable]
# }}}
# {{{ class TriggerBlock:
class TriggerBlock:
    """
    Trigger creating control signal.

    @ivar lineno: Line number of the block in the source file.
    @ivar in_value: Input value.
    @ivar out_value: Output value.

    @ivar policy: Policy of enabling.
    @ivar value: Trigger value.
    @ivar interval: Time interval of the enabled signal.
    @ivar count: Number of hardware samples being enabled.
    """
    def __init__(self, lineno, in_value, out_value, properties):
        self.lineno = lineno
        self.in_value = in_value
        self.out_value = out_value

        self.policy = 'value'
        self.value = 1.0
        self.interval = None
        self.count = 1
        for prop in properties:
            if prop.name == 'policy':
                self.policy = get_identifier_property(prop)
                if self.policy not in ('up', 'down', 'value'):
                    msg = "Trigger policy should be 'up', 'down', or 'value'"
                    report.report_error([self.lineno], msg)
                    # Program dies
            elif prop.name == 'value':
                self.value = get_real_property(prop)
            elif prop.name == 'interval':
                self.interval = get_nonneg_real_property(prop, True)
                self.count = None
            elif prop.name == 'count':
                self.count = get_nonneg_int_property(prop, True)
                self.interval = None
            else:
                report.report_error([prop.lineno], "Unknown property '{}' of trigger block".format(prop.name))

    def get_name(self):
        return 'trigger'

    def get_in_values(self):
        return [self.in_value]

    def get_out_values(self):
        return [self.out_value]

    def set_inout_values(self, in_values, out_values):
        assert len(in_values) == 1
        assert len(out_values) == 1
        self.in_values = in_values
        self.out_values = out_values
# }}}
# {{{ class FilterBlock:
class FilterBlock:
    """
    Filter cleaning up a signal.

    @ivar lineno: Line number of the block in the source file.
    @ivar in_value: Input value.
    @ivar out_value: Output value.

    @ivar weights: Multiplication weights.
    """
    def __init__(self, lineno, in_value, out_value, properties):
        self.lineno = lineno
        self.in_value = in_value
        self.out_value = out_value

        self.weights = [1.0]
        for prop in properties:
            if prop.name == 'weights':
                self.weights = get_reallist_property(prop)
                if len(self.weights) == 0:
                    msg = "Weights parameter needs at least one weight value"
                    report.report_error([self.lineno], msg)
            else:
                report.report_error([prop.lineno], "Unknown property '{}' of filter block".format(prop.name))

    def get_name(self):
        return 'filter'

    def get_in_values(self):
        return [self.in_value]

    def get_out_values(self):
        return [self.out_value]

    def set_inout_values(self, in_values, out_values):
        assert len(in_values) == 1
        assert len(out_values) == 1
        self.in_values = in_values
        self.out_values = out_values
# }}}
# {{{ def make_function_block(lineno, out_value, properties):
def make_function_block(lineno, out_value, properties):
    """
    Construct an appropriate function generator for the requested function.
    """
    shape = None
    for prop in properties:
        if prop.name == 'shape':
            shape = get_identifier_property(prop)
            break

    if shape is None:
        report.report_error([lineno], "Function generator has no 'shape' property")
        # Program dies


    if shape == 'sine':
        return SineFunctionBlock(lineno, out_value, properties)
    if shape == 'block':
        return BlockFunctionBlock(lineno, out_value, properties)
    if shape == 'triangle':
        return TriangleFunctionBlock(lineno, out_value, properties)

    report.report_error([lineno], "Unrecognized function generator shape '{}'".format(shape))
    # Program dies

class BaseFunctionBlock:
    """
    Function generator.

    @ivar lineno: Line number of the block in the source file.
    @ivar out_value: Output value.
    @ivar code_name: Name of the structure in C code.
    """
    def __init__(self, lineno, name, out_value, code_name):
        self.lineno = lineno
        self.name = name
        self.out_value = out_value
        self.code_name = code_name

    def get_name(self):
        return self.name

    def get_in_values(self):
        return []

    def get_out_values(self):
        return [self.out_value]

    def set_inout_values(self, in_values, out_values):
        assert len(in_values) == 0
        assert len(out_values) == 1
        self.out_value = out_values[0]

class SineFunctionBlock(BaseFunctionBlock):
    """
    Sine function generator.

    @ivar frequency: Frequency of the signal.
    @ivar base: Base value of the signal.
    @ivar amplitude: Size of the signal.
    @ivar max_samples: Maximum allowed number of samples.
    """
    def __init__(self, lineno, out_value, properties):
        BaseFunctionBlock.__init__(self, lineno, 'sine_function', out_value, "Sine")
        self.frequency = 1.0
        self.base = 0.0
        self.amplitude = 1.0
        self.max_samples = 100
        for prop in properties:
            if prop.name == 'shape':
                # Already enforced by make_function_block function.
                continue

            if prop.name == 'frequency':
                self.frequency = get_nonneg_real_property(prop, True)
                if self.frequency <= 0:
                    msg = "Non-positive frequency '{}' not supported, value should be bigger than 0"
                    report.report_error([lineno], msg.format(self.frequency))
                    # Program dies
                continue

            if prop.name == 'base':
                self.base = get_real_property(prop)
                continue

            if prop.name == 'amplitude':
                self.amplitude = get_nonneg_real_property(prop)
                continue

            if prop.name == 'max_samples':
                self.max_samples = get_nonneg_int_property(prop, True)
                continue

            report.report_error([prop.lineno], "Unknown property '{}' of sine function".format(prop.name))

class BlockFunctionBlock(BaseFunctionBlock):
    def __init__(self, lineno, out_value, properties):
        BaseFunctionBlock.__init__(self, lineno, 'block_function', out_value, "Block")
        self.frequency = 1.0
        self.duty_cycle = 0.5
        self.low = 0.0
        self.high = 1.0

        for prop in properties:
            if prop.name == 'shape':
                # Already enforced by make_function_block function.
                continue

            if prop.name == 'frequency':
                self.frequency = get_nonneg_real_property(prop, True)
                continue

            if prop.name == 'duty_cycle':
                self.duty_cycle = get_nonneg_real_property(prop)
                continue

            if prop.name == 'low':
                self.low = get_real_property(prop)
                continue

            if prop.name == 'high':
                self.high = get_real_property(prop)
                continue

            report.report_error([prop.lineno], "Unknown property '{}' of block function".format(prop.name))

class TriangleFunctionBlock(BaseFunctionBlock):
    def __init__(self, lineno, out_value, properties):
        BaseFunctionBlock.__init__(self, lineno, 'triangle_function', out_value, "Triangle")
        self.frequency = 1.0
        self.peak_at = 0.5
        self.low = 0.0
        self.high = 1.0

        for prop in properties:
            if prop.name == 'shape':
                # Already enforced by make_function_block function.
                continue

            if prop.name == 'frequency':
                self.frequency = get_nonneg_real_property(prop, True)
                continue

            if prop.name == 'peak_at':
                self.peak_at = get_nonneg_real_property(prop)
                continue

            if prop.name == 'low':
                self.low = get_real_property(prop)
                continue

            if prop.name == 'high':
                self.high = get_real_property(prop)
                continue

            report.report_error([prop.lineno], "Unknown property '{}' of triangle function".format(prop.name))
# }}}

def build_ast(filename):
    """
    Parse the file, and convert parse information to a list of block objects.
    """
    tree = parse_blocks.parse_file(filename)
    ring_index = 0 # First available index number of a ring to or from matlab.

    ast_blocks = []
    for block in tree:
        kind, lineno, out_values, in_values, properties = block
        if kind == 'config':
            assert(len(in_values) == 0)
            assert(len(out_values) == 0)
            ast_blocks.append(Configuration(lineno, properties))
            continue

        if kind == 'from_matlab':
            assert(len(in_values) == 0)
            ast_blocks.append(FromMatlabBlock(lineno, out_values, properties, ring_index))
            ring_index = ring_index + 1
            continue

        if kind == 'to_matlab':
            assert(len(out_values) == 0)
            ast_blocks.append(ToMatlabBlock(lineno, in_values, properties, ring_index))
            ring_index = ring_index + 1
            continue

        if kind == 'trigger':
            assert len(in_values) == 1
            assert len(out_values) == 1
            ast_blocks.append(TriggerBlock(lineno, in_values[0], out_values[0], properties))
            continue

        if kind == 'filter':
            assert len(in_values) == 1
            assert len(out_values) == 1
            ast_blocks.append(FilterBlock(lineno, in_values[0], out_values[0], properties))
            continue

        if kind == 'function':
            assert len(in_values) == 0
            assert len(out_values) == 1
            ast_blocks.append(make_function_block(lineno, out_values[0], properties))
            continue

        print(block)
        assert False # Should not happen

    return ast_blocks


def get_config(blocks):
    """
    Split the blocks in 'configuration block', and 'everything else'.
    """
    config = None
    signal_blocks = []
    for block in blocks:
        if isinstance(block, Configuration):
            if config  is not None:
                msg = "Specification has more than one 'config' block"
                report.report_error([block.lineno, config.lineno], msg)

            config = block
        else:
            signal_blocks.append(block)

    if config is None:
        msg = "Missing configuration block in the specification"
        report.report_error(None, msg)
        # Program dies

    return signal_blocks, config


def check_buf_unbuf_io(lineno, blk_type, buffercount, buffersize, buffered):
    if buffercount is None:
        msg = "{} 'buffercount' property is not set".format(blk_type)
        report.report_error([lineno], msg)

    if buffersize is None:
        msg = "{} 'buffersize' property is not set".format(blk_type)
        report.report_error([lineno], msg)

    if buffered:
        if buffersize == 0:
            msg = "Application runs in buffered mode, but {} 'buffersize' property is 0"
            report.report_error([lineno], msg.format(blk_type))

    else:
        if buffercount != 0 and buffersize != 0:
            msg = ("Application runs in unbuffered mode, but {} 'buffercount' and " +
                   "'buffersize' properties are both non-zero")
            report.report_error([lineno], msg.format(blk_type))


BUF_UNBUF = {False: "unbuffered", True: "buffered"}

def check_setup(blocks, config, buffered):
    """
    Check the setup, and copy default values as needed.

    @param blocks: Blocks to examine.
    @param buffered: Wehether the data acquisition runs in buffered mode.
    """
    if config.buffered != buffered:
        # This currently never happens, since 'buffered' is retrieved from the configuration block.
        msg = "Application runs in {} mode, but configuration 'buffered' property uses {} mode"
        report.report_error([config.lineno], msg.frmat(BUF_UNBUF[buffered], BUF_UNBUF[config.buffered]))

    def asg_default(param_val, default_val):
        if param_val is not None:
            return param_val
        return default_val

    # Copy global size/count defaults to both direction defaults
    config.default_to_bufsize = asg_default(config.default_to_bufsize, config.default_bufsize)
    config.default_from_bufsize = asg_default(config.default_from_bufsize, config.default_bufsize)
    config.default_to_bufcount = asg_default(config.default_to_bufcount, config.default_bufcount)
    config.default_from_bufcount = asg_default(config.default_from_bufcount, config.default_bufcount)

    if not buffered:
        # Set uninitialized defaults to 0 for unbuffered mode.
        config.default_to_bufsize    = asg_default(config.default_to_bufsize, 0)
        config.default_from_bufsize  = asg_default(config.default_from_bufsize, 0)
        config.default_to_bufcount   = asg_default(config.default_to_bufcount, 0)
        config.default_from_bufcount = asg_default(config.default_from_bufcount, 0)

    else:
        # Buffered mode
        config.default_prewrite = asg_default(config.default_prewrite, 0)

    # Copy default buffer settings to all groups if not overridden.
    for blk in blocks:
        if isinstance(blk, FromMatlabBlock):
            blk.buffersize  = asg_default(blk.buffersize,  config.default_from_bufsize)
            blk.buffercount = asg_default(blk.buffercount, config.default_from_bufcount)

        elif isinstance(blk, ToMatlabBlock):
            blk.buffersize = asg_default(blk.buffersize, config.default_to_bufsize)
            blk.buffercount = asg_default(blk.buffercount, config.default_to_bufcount)
            blk.synchronous = asg_default(blk.synchronous, config.default_to_synchronous)
            if buffered:
                blk.prewrite = asg_default(blk.prewrite, config.default_prewrite)

            if blk.enable:
                if blk.synchronous:
                    msg = "to_matlab block using both 'enable' and 'synchronous' may not work as intended"
                    report.report_error([blk.lineno], msg, type='warning')

                if not buffered:
                    msg = "to_matlab block cannot use 'enable' in unbuffered mode"
                    report.report_error([blk.lineno], msg, type='error')


    from_count = 0
    to_count = 0
    for blk in blocks:
        if isinstance(blk, FromMatlabBlock):
            check_buf_unbuf_io(blk.lineno, "from_matlab", blk.buffercount, blk.buffersize, buffered)
            from_count = from_count + 1

        elif isinstance(blk, ToMatlabBlock):
            check_buf_unbuf_io(blk.lineno, "to_matlab", blk.buffercount, blk.buffersize, buffered)
            to_count = to_count + 1

            if buffered:
                # Prewrite checks (buffered mode only).
                if blk.prewrite > blk.buffersize - 1:
                    msg = ("to_matlab 'prewrite' property must be less than the buffer " +
                           "size (max is {}, found {})").format(blk.buffersize - 1, blk.prewrite)
                    report.report_error([blk.lineno], msg)

            elif blk.prewrite is not None:
                # Unbuffered mode, prewrite makes no sense.
                msg = "to_matlab 'prewrite' property cannot be used in unbuffered mode"
                report.report_error([blk.lineno], msg)

    if not buffered and from_count > 1:
        msg = "Application runs in unbuffered mode, at most one from_matlab block is allowed"
        report.report_error(None, msg)

    if not buffered and to_count > 1:
        msg = "Application runs in unbuffered mode, at most one to_matlab block is allowed"
        report.report_error(None, msg)


def add_block_to_signal(signame, index, block, signals):
    """
    Add a block to the table of signals.

    @param signame: Name of the signal to attach to.
    @param index: Whether the signal is input (0) or output (1).
    @param block: Block to add.
    @param signals: Signals to blocks mapping that is being updated.
    """
    sig_io = signals.get(signame)
    if sig_io is None:
        sig_io = ([], [])
        signals[signame] = sig_io

    sig_io[index].append(block)


def build_graph(blocks):
    """
    Build a dependency graph, and perform sanity checking of the connections

    @return: The graph, as readers and writers of each edge.
    """
    signals = {} # Mapping of value name to ([read-blocks],[write-blocks])

    # Collect connections between blocks over signals.
    for block in blocks:
        for signal in block.get_in_values():
            add_block_to_signal(signal, 0, block, signals)
        for signal in block.get_out_values():
            add_block_to_signal(signal, 1, block, signals)

    # Check names of the signals
    names.check_names(list(signals.keys()))

    # Add implicit Soem blocks for the hardware signals.
    in_signals  = [sig for sig in sorted(signals.keys()) if sig.get_inputkind() is not None]
    if len(in_signals) > 0:
        block = SoemInputs(None, in_signals)
        blocks.append(block)
        for signal in in_signals:
            add_block_to_signal(signal, 1, block, signals)

    out_signals = [sig for sig in sorted(signals.keys()) if sig.get_outputkind() is not None]
    if len(out_signals) > 0:
        block = SoemOutputs(None, out_signals)
        blocks.append(block)
        for signal in in_signals:
            add_block_to_signal(signal, 0, block, signals)

    # Check connections
    found_error = False
    for signal, (read_blocks, write_blocks) in signals.items():
        if len(write_blocks) == 0:
            assert len(read_blocks) > 0 # Else the signal would not exist.
            msg = "Signal '{}' must have a block writing it (none found)"
            msg = msg.format(signal)
            report.report_error([read_blocks[0].lineno], msg, keep_running = True)
            found_error = True

        # But not too often.
        elif len(write_blocks) > 1:
            msg = "Signal '{}' has more than one block writing it ({} found)"
            msg = msg.format(signal, len(write_blocks))
            lines = [w.lineno for w in write_blocks if w.lineno is not None]
            report.report_error(lines, msg, keep_running = True)
            found_error = True

        # For non-output signals, reading them is useful.
        if signal.get_outputkind() is None and len(read_blocks) == 0:
            assert len(write_blocks) > 0 # Else the signal would not exist.
            msg = "Signal '{}' is written but not used".format(signal)
            report.report_error([write_blocks[0].lineno], msg, type='warning')

    if found_error:
        sys.exit(1)

    return signals

def dump_graph(signals):
    """
    Dump the constructed graph to the output.
    """
    for signal, (inblocks, outblocks) in signals.items():
        sig_txt = str(signal)
        print("{} <- {}".format(sig_txt, ", ".join(b.get_name() for b in outblocks)))
        print("{} -> {}".format(" "*len(sig_txt), ", ".join(b.get_name() for b in inblocks)))
        print("")

def get_useless_blocks(blocks, edges):
    """
    Find blocks that have non-used output (recursively).

    @return: useless signals and useless blocks.
    """
    useless_blocks = set()
    useless_signals = set()

    while True:
        progress = False

        # Discard unused signals.
        for signal, (read_blocks, write_blocks) in edges.items():
            if signal in useless_signals:
                # Already discarded, do not consider again.
                continue

            if any(bl not in useless_blocks for bl in read_blocks):
                # Signal is being read.
                continue

            if signal.get_outputkind() is not None:
                # It's an output signal.
                continue

            msg = "Discarding signal '{}'".format(signal)
            report.report_error([signal.lineno], msg, type='info')
            useless_signals.add(signal)
            progress = True

        for block in blocks:
            if isinstance(block, (ToMatlabBlock, SoemOutputs)):
                # ToMatlab blocks are always useful.
                continue

            if block in useless_blocks:
                # Already discarded the block.
                continue

            if any(sig not in useless_signals for sig in block.get_out_values()):
                # Some signal is still used.
                continue

            msg = "Discarding '{}' block".format(block.get_name())
            report.report_error([block.lineno], msg, type='info')
            useless_blocks.add(block)
            progress = True

        if not progress:
            break

    return useless_signals, useless_blocks

def cleanup_blocks_signals(blocks, edges):
    """
    Remove the useless blocks and edges from the graph
    """
    useless_signals, useless_blocks = get_useless_blocks(blocks, edges)

    if len(useless_signals) == 0 and len(useless_blocks) == 0:
        return blocks, edges

    good_blocks = []
    for block in blocks:
        if block in useless_blocks:
            continue

        good_in_values  = [sig for sig in block.get_in_values()  if sig not in useless_signals]
        good_out_values = [sig for sig in block.get_out_values() if sig not in useless_signals]
        block.set_inout_values(good_in_values, good_out_values)
        good_blocks.append(block)

    good_edges = {}
    for signal, (read_blocks, write_blocks) in edges.items():
        if signal in useless_signals:
            continue

        read_blocks  = [bl for bl in read_blocks  if bl not in useless_blocks]
        write_blocks = [bl for bl in write_blocks if bl not in useless_blocks]
        good_edges[signal] = (read_blocks, write_blocks)

    return good_blocks, good_edges

def decide_block_computes(blocks, edges):
    """
    Decide in which order to compute the blocks such that signals are written before being read.
    """
    # Repeatedly find a block that does not read invalid signals.
    valid_signals = set()
    done_blocks = set()

    # Find blocks until all done or no block with only valid inputs.
    # XXX Pretty stupid implementation, but fast enough for small networks.
    ordered_blocks = []
    while len(done_blocks) < len(blocks):
        valid_blocks = []
        for block in blocks:
            if block in done_blocks:
                continue

            if any(sig not in valid_signals for sig in block.get_in_values()):
                continue

            valid_blocks.append(block)

        if len(valid_blocks) == 0: # Cycle found!
            invalid_blk_names = set()
            invalid_blk_lines = []
            for block in blocks:
                if block in done_blocks:
                    continue

                invalid_blk_names.add("\'{}\'".format(block.get_name))
                invalid_blk_lines.append(block.lineno)
            msg = "Signal cycle detected in the following blocks: {}"
            msg = msg.format(", ".join(sorted(invalid_blk_names)))
            report.report_error(invalid_blk_lines, msg)


        # If we can do Soem input, do it asap.
        for block in valid_blocks:
            if isinstance(block, SoemInputs):
                valid_blocks = [block]
                break

        for block in valid_blocks:
            ordered_blocks.append(block)

            done_blocks.add(block)
            for sig in block.get_out_values():
                assert sig not in valid_signals
                valid_signals.add(sig)

        # And let's try again.

    return ordered_blocks

