/** buffer.c Handling buffers and ring buffers. */

#include "stdincludes.h"

#include <pthread.h>
#include <string.h>
#include "buffer.h"

/**
 * Initialize a buffer structure.
 * @param buf Buffer to initialize.
 * @param data Sample data base address.
 * @param width Number of columns in the buffer.
 * @param height Number of rows in the buffer.
 */
void InitBuffer(Buffer *buf, double *data, uint32_t width, uint16_t height)
{
    buf->width = width;
    buf->start = 0;
    buf->used = 0;
    buf->height = height;
    buf->data = data;
}

/**
 * Fill a buffer from the matrix data in \a src_values.
 * @param dest_buf Buffer to write to.
 * @param src_values Values to write to the buffer.
 * @param num_rows Number of rows in the matrix data. Must be equal to the number of rows in the buffer.
 * @param num_cols Number of columns in the matrix data,
 * @return Number of *uncopied* columns in the \a src_values (that is, \c 0 is good, non-\c 0 is bad).
 */
uint32_t BufferBlockFill(Buffer *dest_buf, const double *src_values, uint16_t num_rows, uint32_t num_cols)
{
    uint32_t copied_columns; /* Total number of copied columns. */

    assert(dest_buf->height == num_rows); /* Number of rows must match. */

    copied_columns = (num_cols > dest_buf->width) ? dest_buf->width : num_cols;
    memcpy(dest_buf->data, src_values, sizeof(double) * copied_columns * num_rows);
    dest_buf->start = 0;
    dest_buf->used = copied_columns;

    return num_cols - copied_columns;
}

/**
 * Mass-copy collected data from the buffer to the matrix data-part.
 * @param dest_values Base address to write the values to.
 * @param num_rows Number of rows in the \a dest_values. Must match the number of rows in the buffer.
 * @param num_cols Number of columns in the \a dest_values. It is recommended to make it equal to the
 *      number of used columns in the buffer.
 * @param src_buf Source of the data, buffer is cleared afterwards.
 * @return Number of columns *not* copied (that is, \c 0 is good, non-\c 0 is bad).
 * @note Superfluous columns in the \a dest_values are not initialized.
 */
uint32_t BufferBlockEmpty(double *dest_values, uint16_t num_rows, uint32_t num_cols, Buffer *src_buf)
{
    uint32_t copied_columns; /* Total number of copied columns. */
    uint32_t copy_width; /* Number of columns copied in a single memcpy. */
    uint32_t remaining_cols; /* Remaining number of columns in \a src_buf. */

    assert(src_buf->height == num_rows); /* Number of rows must match. */

    copied_columns = (num_cols > src_buf->used) ? src_buf->used : num_cols;
    remaining_cols = src_buf->used - copied_columns;

    /* Copy block from 'start' column to end of buffer. */
    copy_width = src_buf->width - src_buf->start;
    if (copy_width > copied_columns) copy_width = copied_columns;
    if (copy_width > 0) {
        memcpy(dest_values, src_buf->data + (src_buf->start * num_rows), sizeof(double) * copy_width * num_rows);
        copied_columns -= copy_width;
        dest_values += copy_width * num_rows;
    }
    /* Copy remaining columns between 0 and 'start' column. */
    if (copied_columns > 0) {
        memcpy(dest_values, src_buf->data, sizeof(double) * copied_columns * num_rows);
        dest_values += copied_columns * num_rows;
    }

    /* XXX To recover the uncopied columns, data must be shifted here to column 0, or
     * 'start' and 'used' must be updated. For now, just destroy remaining columns. */
    src_buf->start = 0;
    src_buf->used = 0;

    return remaining_cols;
}

/**
 * Initialize a read buffer from scratch.
 * @param buf_read Read buffer to clear.
 * @param pitch Number of values to skip to get at the next column.
 */
void InitBufferRead(BufferRead *buf_read, size_t pitch)
{
    buf_read->buffer = NULL;
    buf_read->data_first = NULL;
    buf_read->data_end = NULL;
    buf_read->data_current = NULL;
    buf_read->pitch = pitch;
    buf_read->used = 0;
}

/**
 * Setup a read buffer for reading from a new buffer.
 * @param buf_read Read buffer to initialize.
 * @param buffer Bufer to use for reading.
 * @note Buffer may not be empty.
 */
void SetupBufferRead(BufferRead *buf_read, Buffer *buffer)
{
    assert(buffer->height == buf_read->pitch);
    assert(buffer->used > 0);
    assert(buffer->start < buffer->width);

    buf_read->buffer = buffer;
    buf_read->data_first = buffer->data;
    buf_read->data_end = buf_read->data_first + (buf_read->pitch * buffer->width);
    buf_read->used = buffer->used;
    buf_read->data_current = buf_read->data_first + (buf_read->pitch * buffer->start);
}

/**
 * Get a column of data from the buffer.
 * @param buf_read Read buffer to use.
 * @return Column of data to use, or \c NULL.
 * @note This function will repeatedly return the last column of data.
 *      The #BufferReadHasUnread function says whether the buffer should be replaced.
 */
double *ReadFromBuffer(BufferRead *buf_read)
{
    assert(BufferReadHasBuffer(buf_read));

    double *retval = buf_read->data_current;
    if (retval == NULL || buf_read->used == 0) return NULL;

    buf_read->data_current += buf_read->pitch;
    if (buf_read->data_current == buf_read->data_end) buf_read->data_current = buf_read->data_first;
    buf_read->used--;
    return retval;
}

/**
 * Reset the buffer to empty state, and disconnect it from the read buffer.
 * @param buf_read Read buffer to finish.
 */
void FinishBufferRead(BufferRead *buf_read)
{
    Buffer *buf = buf_read->buffer;
    if (buf != NULL) {
        buf->start = 0;
        buf->used = 0;
    }

    buf_read->buffer = NULL;
    buf_read->data_first = NULL;
    buf_read->data_end = NULL;
    buf_read->data_current = NULL;
    /* pitch preserved. */
    buf_read->used = 0;
}

/**
 * Initialize the buffer write structure.
 * @param buf_write Buffer write structure to initialize.
 * @param pitch Number of values to skip to jump to the next column in the data.
 * @param prewrite Number of samples to save before receiving the first proper measurement.
 */
void InitBufferWrite(BufferWrite *buf_write, size_t pitch, uint32_t prewrite)
{
    buf_write->buffer = NULL;
    buf_write->data_first = NULL;
    buf_write->data_end = NULL;
    buf_write->data_current = NULL;
    buf_write->pitch = pitch;
    buf_write->width = 0;
    buf_write->used = 0;
    buf_write->write_count = 0;
    buf_write->start = 0;
    buf_write->prewrite = prewrite;
}

/**
 * Setup the buffer write structure for writing to the given \a buffer.
 * @param buf_write Buffer write structure to set up.
 * @param buffer to use for storing sample data.
 * @todo Check that prewrite upper limit is less then width of a buffer.
 */
void SetupBufferWrite(BufferWrite *buf_write, Buffer *buffer)
{
    assert(buffer->height == buf_write->pitch);
    assert(buf_write->prewrite < buffer->width);

    buf_write->buffer = buffer;
    buf_write->width = buffer->width;
    buf_write->data_first = buffer->data;

    buf_write->data_current = buf_write->data_first;
    buf_write->data_end = buf_write->data_first + (buf_write->pitch * buf_write->width);
    buf_write->used = 0;
    buf_write->write_count = 0;
    buf_write->start = 0;
}

/**
 * Prewrite a sample to the buffer.
 * param buf_write Buffer writer to use.
 * @return Start address for prewriting the column of samples.
 * @pre Buffer writer must have a buffer, and it must not contain real samples.
 */
double *BufferWritePrewrite(BufferWrite *buf_write)
{
    double *address;

    assert(BufferWriteHasBuffer(buf_write) && !BufferWriteHasData(buf_write));

    if (buf_write->prewrite == 0) {
        /* No prewriting is a simple case. */
        return buf_write->data_current;
    }
    /* else, store the new sample after the previously written samples. */

    if (buf_write->write_count < buf_write->prewrite) buf_write->write_count++;

    address = buf_write->data_current;
    buf_write->data_current = buf_write->data_current + buf_write->pitch;
    if (buf_write->data_current == buf_write->data_end) buf_write->data_current = buf_write->data_first;
    return address;
}

/**
 * Write a proper sample to the buffer.
 * param buf_write Buffer writer to use.
 * @return Start address for writing the column of samples.
 * @pre Buffer writer must have a buffer, and it must have space for real samples.
 */
double *BufferWriteReal(BufferWrite *buf_write)
{
    double *address;

    assert(BufferWriteHasBuffer(buf_write) && BufferWriteHasFreeSpace(buf_write));

    if (buf_write->used == 0) {
        /* Switching from prewriting to real writing. */
        buf_write->used = buf_write->write_count + 1; /* +1 for the first real data. */

        /* Compute 'start' from 'data_current'. */
        size_t index = (buf_write->data_current - buf_write->data_first) / buf_write->pitch;
        assert(index < buf_write->width);
        if (index >= buf_write->write_count) {
            /* Last 'write_count' samples are in one block before 'data_current'. */
            buf_write->start = index - buf_write->write_count;
        } else {
            /* Last 'write_count' samples wraps around backwards.' */
            size_t remaining = buf_write->write_count - index;
            buf_write->start = buf_write->width - remaining;
            assert(buf_write->start < buf_write->width);
        }
    } else {
        buf_write->used++;
    }

    address = buf_write->data_current;
    buf_write->data_current = buf_write->data_current + buf_write->pitch;
    if (buf_write->data_current == buf_write->data_end) buf_write->data_current = buf_write->data_first;

    return address;
}

/**
 * Finish writing to the currently attached buffer by detaching it.
 * @param buf_write Buffer writer to clean.
 */
void FinishBufferWrite(BufferWrite *buf_write)
{
    if (buf_write->buffer != NULL) {
        Buffer *buf = buf_write->buffer;
        if (BufferWriteHasData(buf_write)) {
            buf->start = buf_write->start;
            buf->used = buf_write->used;
        } else {
            /* Nothing has been written in it (except perhaps some prewrite data). */
            buf->start = 0;
            buf->used = 0;
        }
    }

    buf_write->buffer = NULL;
    buf_write->data_first = NULL;
    buf_write->data_end = NULL;
    buf_write->data_current = NULL;
    /* buf_write->pitch saved. */
    buf_write->width = 0;
    buf_write->used = 0;
    buf_write->write_count = 0;
    buf_write->start = 0;
    /* buf_write->prewrite saved. */
}

/**
 * Initialize a ring buffer.
 * @param ring Ring buffer to initialize.
 * @param buffers First buffer available for use by the ring buffer.
 * @param num_buffers Number of buffers available for use.
 * @param parent_buffers If set, assign the buffers to the parent stage, else to the process stage.
 */
void InitRingbuffer(Ringbuffer *ring, Buffer *buffers, uint16_t num_buffers, BOOL parent_buffers)
{
    pthread_mutex_init(&ring->lock, NULL);
    ring->buffers = buffers;
    ring->num_buffers = num_buffers;

    ring->parent_index = 0;
    ring->outbuffer_index = 0;
    ring->process_index = 0;
    ring->inbuffer_index = 0;

    ring->parent_count = parent_buffers ? num_buffers : 0;
    ring->outbuffer_count = 0;
    ring->process_count = parent_buffers ? 0 : num_buffers;
    ring->inbuffer_count = 0;
}

/**
 * Destroy the data of a ring buffer. */
void DestroyRingbuffer(Ringbuffer *ring)
{
    pthread_mutex_destroy(&ring->lock);
    ring->num_buffers = 0;
    ring->buffers = NULL;

    ring->parent_index = 0;
    ring->outbuffer_index = 0;
    ring->process_index = 0;
    ring->inbuffer_index = 0;

    ring->parent_count = 0;
    ring->outbuffer_count = 0;
    ring->process_count = 0;
    ring->inbuffer_count = 0;
}

/*
 * /index             /index        /index     /index
 * V parent           V outbuffer   V process  V inbuffer
 * | A                | B           | C        | D            |
 * |__________________|_____________|__________|______________|
 *
 *     /index             /index        /index     /index
 *     V parent           V outbuffer   V process  V inbuffer
 *   A |                B |           C |        D |
 * ____|__________________|_____________|__________|___________
 */

/* Update parent (1 done):
 * parent -> inbuffer (1 buffer)
 *
 * parent_index += 1
 * parent_count -= 1
 * in_count += 1
 *
 * outbuffer -> parent (all buffers)
 *
 * parent_count += out_count
 * out_index += out_count
 * out_count = 0
 * # In the picture, outbuffer would not have any buffers after this.
 */

/**
 * Update the buffers around the parent thread.
 * @param ring Ring buffer to update.
 * @param num_done Number of buffers of the parent stage considered done.
 * @return Updated number of buffers owned by parent stage.
 * @note Code does not check value of \a num_done.
 */
uint16_t UpdateParentRing(Ringbuffer *ring, uint16_t num_done)
{
    pthread_mutex_lock(&ring->lock);

    ring->parent_index += num_done;
    if (ring->parent_index >= ring->num_buffers) ring->parent_index -= ring->num_buffers;
    ring->parent_count = ring->parent_count - num_done + ring->outbuffer_count;
    ring->inbuffer_count += num_done;
    ring->outbuffer_index += ring->outbuffer_count;
    if (ring->outbuffer_index >= ring->num_buffers) ring->outbuffer_index -= ring->num_buffers;
    ring->outbuffer_count = 0;

    pthread_mutex_unlock(&ring->lock);

    return ring->parent_count;
}

/**
 * Update ring administration at the parent side.
 * @param ring Ring buffer to update.
 * @param num_done Number of buffers of the process stage considered done.
 * @note Code assumes the caller has locked the ring for exclusive access.
 */
static void RawUpdateProcessRing(Ringbuffer *ring, uint16_t num_done)
{
    /* ## ASSUMES RING IS LOCKED BY CALLER ## */

    ring->process_index += num_done;
    if (ring->process_index >= ring->num_buffers) ring->process_index -= ring->num_buffers;
    ring->process_count = ring->process_count - num_done + ring->inbuffer_count;
    ring->outbuffer_count += num_done;
    ring->inbuffer_index += ring->inbuffer_count;
    if (ring->inbuffer_index >= ring->num_buffers) ring->inbuffer_index -= ring->num_buffers;
    ring->inbuffer_count = 0;
}

/**
 * Update the buffers around the process thread.
 * @param ring Ring buffer to update.
 * @param num_done Number of buffers of the process stage considered done.
 * @return Updated number of buffers owned by process stage.
 * @note Code does not check value of \a num_done against buffers that exist in that stage.
 */
uint16_t UpdateProcessRing(Ringbuffer *ring, uint16_t num_done)
{
    pthread_mutex_lock(&ring->lock);

    RawUpdateProcessRing(ring, num_done);

    pthread_mutex_unlock(&ring->lock);
    return ring->process_count;
}

/**
 * Get oldest parent buffer.
 * @param ring Ring buffer to query.
 * @return Oldest buffer in parent phase, or \c NULL.
 * @todo Add interface for getting more parent buffers.
 */
Buffer *GetParentBuffer(Ringbuffer *ring)
{
    /* Parent phase can be access lock-free. */
    if (ring->parent_count == 0) return NULL;
    return ring->buffers + ring->parent_index;
}

/**
 * Get oldest process buffer.
 * @param ring Ring buffer to query.
 * @return Oldest buffer in process phase, or \c NULL.
 * @todo Add interface for getting more process buffers.
 */
Buffer *GetProcessBuffer(Ringbuffer *ring)
{
    /* Process phase can be access lock-free. */
    if (ring->process_count == 0) return NULL;
    return ring->buffers + ring->process_index;
}

/**
 * Get the base addres of a column in the current buffer for reading buffer data.
 * @param buf_read Read-buffer administration.
 * @param ring Ring buffer to get buffers from.
 * @param [out] error Statically allocated error message, if any. An error might
 *      be written even if it is recoverable, in which case it might be treated
 *      as a warning rather than a fatal error.
 * @return Base address of the requested column, or \c NULL if no column address could be obtained.
 */
double *GetReadColumn(BufferRead *buf_read, Ringbuffer *ring, const char **error)
{
    /* If bufread has no buffer, supply one if possible. */
    if (!BufferReadHasBuffer(buf_read)) {
        if (GetProcessBuffer(ring) == NULL && UpdateProcessRing(ring, 0) == 0) {
            /* No buffer! */
            if (error != NULL) *error = "No Matlab buffer available for reading";
            return NULL;
        }

        SetupBufferRead(buf_read, GetProcessBuffer(ring));
    } else {
        /* Have a buffer, but does it have unread data? */
        if (!BufferReadHasUnread(buf_read)) {
            FinishBufferRead(buf_read);
            if (UpdateProcessRing(ring, 1) == 0) {
                /* No unread data, and no new buffer, we're in trouble! */
                if (error != NULL) *error = "No Matlab buffer available for reading";
                return NULL;
            }
            SetupBufferRead(buf_read, GetProcessBuffer(ring));
        }
    }

    /* Use whatever we have at this point. */
    if (error != NULL) *error = NULL;
    return ReadFromBuffer(buf_read);
}

/**
 * Get the base addres of a column in the current buffer for writing buffer data.
 * @param buf_write Write-buffer administration.
 * @param ring Ring buffer to get buffers from.
 * @param prewrite If set, the new value is prewritten, else it's a 'real' data value.
 * @param [out] error Statically allocated error message, if any. An error might
 *      be returned even if a valid base address is returned, which should be
 *      interpreted as a warning.
 * @return Base address of a column to write to, if not \c NULL.
 */
double *GetWriteColumn(BufferWrite *buf_write, Ringbuffer *ring, BOOL prewrite, const char **error)
{
    if (!BufferWriteHasBuffer(buf_write)) {
        if (GetProcessBuffer(ring) == NULL && UpdateProcessRing(ring, 0) == 0) {
            /* No buffer! */
            if (!prewrite && error != NULL) *error = "No matlab buffer available for writing";
            return NULL;
        }
        SetupBufferWrite(buf_write, GetProcessBuffer(ring));
    } else {
        if (!BufferWriteHasFreeSpace(buf_write)) {
            FinishBufferWrite(buf_write);
            if (UpdateProcessRing(ring, 1) == 0) {
                if (!prewrite && error != NULL) *error = "No new matlab buffer available for writing";
                return NULL;
            }
            SetupBufferWrite(buf_write, GetProcessBuffer(ring));
        }
    }

    assert(BufferWriteHasFreeSpace(buf_write));
    if (error != NULL) *error = NULL;
    return prewrite ? BufferWritePrewrite(buf_write) : BufferWriteReal(buf_write);
}
