/** @file buffer.h Data buffering and passing between matlab and real-time threads. */

#ifndef BUFFERS_H
#define BUFFERS_H

/**
 * Buffer with data. Structure is only used for storage, reading or writing is
 * done by setting up a #BufferRead or #BufferWrite structure (and afterwards updating the buffer).
 */
struct BufferStruct {
    uint32_t width;  /**< Number of sample columns in the buffer (fixed value). */
    uint32_t start;  /**< Number of columns before the the first sample column in the buffer. */
    uint32_t used;   /**< Number of filled columns in the buffer. */
    uint16_t height; /**< Number of values in a column (fixed value). */
    double *data;    /**< Data store of the buffer, column-wise storage (fixed value). */
};
typedef struct BufferStruct Buffer;

void InitBuffer(Buffer *buf, double *data, uint32_t width, uint16_t height);

/**
 * Get the number of used columns of a buffer.
 * @param buf Buffer to query.
 * @return Number of used columns in the buffer.
 */
#define GetBufferUsedColumns(buf) ((buf)->used)

/**
 * Get the number of available columns in a buffer.
 * @param buf Buffer to query.
 * @return Number of columns that can be stored in the buffer.
 */
#define GetBufferWidth(buf) ((buf)->width)

uint32_t BufferBlockFill(Buffer *dest_buf, const double *src_values, uint16_t num_rows, uint32_t num_cols);
uint32_t BufferBlockEmpty(double *dest_values, uint16_t num_rows, uint32_t num_cols, Buffer *src_buf);

/**
 * Data structure for reading a buffer.
 * - If #buffer is \c NULL, there is no data. Otherwise, #data_first, #data_end, #data_current are valid.
 * - If #used is \c 0, there is no unread column of data. In that case, #data_current points at the
 *   last column, and does not move.
 * - Else, #data_current points at the first unread column of data.
 */
struct BufferReadStruct {
    Buffer *buffer;       /**< Buffer parent structure. */
    double *data_first;   /**< Start of first data column. */
    double *data_end;     /**< First address BEHIND the data. */
    double *data_current; /**< Address of the first available column to read. */
    size_t pitch;         /**< Jump size to the next column (fixed value). */
    uint32_t used;        /**< Number of used columns. */
};
typedef struct BufferReadStruct BufferRead;

/**
 * Check whether the buffer reader has a buffer to read from.
 * @param bufread Read buffer to check.
 * @return Whether a buffer is attached to the buffer read structure.
 */
#define BufferReadHasBuffer(bufread) ((bufread)->buffer != NULL)

/**
 * Does the non-clear #BufferRead structure have unread data?
 * @param bufread Read buffer to check.
 * @return Whether there is unread data in the read buffer.
 * @note Code fails if it has no attached buffer.
 */
#define BufferReadHasUnread(bufread) ((bufread)->used > 0)

void InitBufferRead(BufferRead *buf_read, size_t pitch);
void SetupBufferRead(BufferRead *buf_read, Buffer *buffer);
double *ReadFromBuffer(BufferRead *buf_read);
void FinishBufferRead(BufferRead *buf_read);

/**
 * Data structure for writing a buffer.
 * - If #buffer is \c NULL, there is no data. Otherwise, #data_first, #data_end, #data_current are valid.
 * - If #used is \c 0, there is no proper data collected yet (that is, measurements that need to be kept).
 *   In that case, the buffer is storing prewrite samples.
 *   - #write_count contains the number of prewritten samples (upto #prewrite).
 * - If #used is larger than \c 0, there is proper data in the buffer. Prewriting is not allowed anymore.
 *   - #start points at the first column with data (including relevant.prewritten data).
 *   - #used is the number of samples written so far (including relevant prewritten data).
 */
struct BufferWriteStruct {
    Buffer *buffer;       /**< Buffer parent structure. */
    double *data_first;   /**< Start of first data column. */
    double *data_end;     /**< First address BEHIND the data. */
    double *data_current; /**< Address of the first available column to write. */
    size_t pitch;         /**< Jump size to the next column (fixed value). */
    uint32_t width;       /**< Number of samples that fits in the buffer. */
    uint32_t used;        /**< Number of used columns. */
    uint32_t write_count; /**< Number of written columns (during prewriting). */
    uint32_t start;       /**< First column with real data. */
    uint32_t prewrite;    /**< Number of needed values for prewriting (fixed value). */
};
typedef struct BufferWriteStruct BufferWrite;

/**
 * Check whether the buffer writer has a buffer to write to.
 * @param bufwrite Write buffer to check.
 * @return Whether a buffer is attached to the buffer write structure.
 */
#define BufferWriteHasBuffer(bufwrite) ((bufwrite)->buffer != NULL)

/**
 * Test whether the buffer writer has written proper measurement data to the buffer.
 * @param bufwrite Buffer writer to test.
 * @return Whether real measurement data exists in the buffer.
 */
#define BufferWriteHasData(bufwrite) ((bufwrite)->used > 0)

/**
 * Test whether there is room for storing real samples in the buffer.
 * @param bufwrite Buffer writer to test.
 * @return Whether the buffer still has free space.
 */
#define BufferWriteHasFreeSpace(bufwrite) ((bufwrite)->used < (bufwrite)->width)


void InitBufferWrite(BufferWrite *buf_write, size_t pitch, uint32_t prewrite);
void SetupBufferWrite(BufferWrite *buf_write, Buffer *buffer);
double *BufferWritePrewrite(BufferWrite *buf_write);
double *BufferWriteReal(BufferWrite *buf_write);
void FinishBufferWrite(BufferWrite *buf_write);

/**
 * A ring buffer exchanges data between Matlab parent thread and real-time
 * acquisition thread through buffers.
 * A buffer cycles around in 4 stages.
 * - 'parent'. The buffer is exclusively owned by the Matlab parent thread.
 * - 'inbuffer' The buffer is waiting for its turn to be processed by the
 *   real-time thread.
 * - 'process' The buffer is exclusively owned and processed by the real time
 *   thread.
 * - 'outbuffer' The buffer is waiting for its turn to be processed by the
 *   Matlab parent thread.
 *
 * Each stage has a first buffer (its index), and a length (its count). Note
 * that index+count can wrap in the 'buffers' array. When transferring
 * a buffer, the last buffer of a stage is always taken, and it's added as
 * first buffer of the next stage. Physically it doesn't move, only indices
 * and counts are updated.
 *
 * Only inbuffer and outbuffer is shared (both their buffers and their stage
 * index and count), and needs a lock before it can be accessed.
 *
 * The Matlab thread
 * - takes last entries from the 'outbuffer' stage, and adds them to the front
 *   of its own 'parent' stage, or it
 * - takes last entries from its 'parent' stage, and adds them to the front of
 *   the 'inbuffer' stage.
 *
 * The real-time thread
 * - takes last entries from the 'inbuffer' stage, and adds them to the front
 *   of its own 'process' stage, or it
 * - takes last entries from its 'process' stage, and adds them to the front of
 *   the 'outbuffer' stage.
 *
 * From the Matlab thread, 'outbuffer' doesn't get smaller except by its own
 * actions. From the real-time thread, 'inbuffer' doesn't get smaller except by
 * its own actions.
 */
struct RingbufferStruct {
    pthread_mutex_t lock; /**< Lock for modifying indices and counters. */
    uint16_t num_buffers; /**< Number of buffers in the ring. */
    Buffer *buffers;      /**< Pointer to the first buffer. */

    uint16_t parent_index; /**< Oldest buffer in the parent phase. */
    uint16_t parent_count; /**< Number of buffers in the parent phase. */

    uint16_t outbuffer_index; /**< Oldest buffer in the outbuffer phase. */
    uint16_t outbuffer_count; /**< Number of buffers in the outbuffer phase. */

    uint16_t process_index; /**< Oldest buffer in the process phase. */
    uint16_t process_count; /**< Number of buffers in the process phase. */

    uint16_t inbuffer_index; /**< Oldest buffer in the inbuffer phase. */
    uint16_t inbuffer_count; /**< Number of buffers in the inbuffer phase. */
};
typedef struct RingbufferStruct Ringbuffer;


void InitRingbuffer(Ringbuffer *ring, Buffer *buffers, uint16_t num_buffers, BOOL parent_buffers);
void DestroyRingbuffer(Ringbuffer *ring);

Buffer *GetParentBuffer(Ringbuffer *ring);
Buffer *GetProcessBuffer(Ringbuffer *ring);

uint16_t UpdateParentRing(Ringbuffer *ring, uint16_t num_done);
uint16_t UpdateProcessRing(Ringbuffer *ring, uint16_t num_done);
double *GetReadColumn(BufferRead *buf_read, Ringbuffer *ring, const char **error);
double *GetWriteColumn(BufferWrite *buf_write, Ringbuffer *ring, BOOL prewrite, const char **error);

#endif
