/**
 * @file dbg_test_nanotime.c
 * Code for debugging and testing timing of the system, is not needed for normal operation.
 *
 * Define DBG_NANOTIME to include this code in the program.
 */
#include "stdincludes.h"
#include "nanotime.h"

/* Define DBG_NANOTIME to include this code in the program. */
#ifdef DBG_NANOTIME

static uint64_t base_secs = 0; ///< Base-time in seconds to compare against, for output purposes.

/**
 * Get relative time against a base time in seconds, for output purposes.
 * @param nanos Time to output.
 * @return Time against the base time, in seconds.
 * @note Base time is truncated to full seconds, so relative time does not start at exactly \c 0.0 .
 */
double GetRelativeTime(const NanoTime *nanos)
{
    double rel_time = nanos->tv_nsec / (double)ONE_SEC;

    if (base_secs == 0) base_secs = nanos->tv_sec;

    if (nanos->tv_sec < base_secs) {
        rel_time -= base_secs - nanos->tv_sec;
    } else {
        rel_time += nanos->tv_sec - base_secs;
    }
    return rel_time;
}

/**
 * Get an easy-to-read relative time.
 * @return The current relative time.
 * @note Relative time does not start at exactly \c 0.0 .
 */
double GetRelativeCurrent(void)
{
    NanoTime now = NowNanos();
    return GetRelativeTime(&now);
}

#endif

/* Define TEST_NANOTIME to include the test code. Note that it also defines a main function!
 *
 * Compile:  gcc -DTEST_NANOTIME -Wall -o dbg_test_nanotime dbg_test_nanotime.c nanotime.c
 */
#ifdef TEST_NANOTIME

static BOOL CheckCmp(const NanoTime *t1, const NanoTime *t2, int val)
{
    BOOL ok1, ok2;
    if (val == 0) {
        ok1 = CmpNanos(t1, t2) == 0;
        ok2 = ok1;
    } else if (val > 0) {
        ok1 = CmpNanos(t1, t2) > 0;
        ok2 = CmpNanos(t2, t1) < 0;
    } else {
        ok1 = FALSE; /* Should never happen, but at least it will be noticed. */
        ok2 = FALSE;
    }

    printf("Compare [%lu, %lu] with [%lu, %lu] -> %s\n", t1->tv_sec, t1->tv_nsec, t2->tv_sec, t2->tv_nsec, ok1 ? "ok" : "FAIL");
    printf("Compare [%lu, %lu] with [%lu, %lu] -> %s\n", t2->tv_sec, t2->tv_nsec, t1->tv_sec, t1->tv_nsec, ok2 ? "ok" : "FAIL");
    printf("\n");
    return ok1 & ok2;
}

static BOOL TestCompare(void)
{
    BOOL ok = TRUE;
    NanoTime t_1_50 = MakeNanos(1, 50);
    NanoTime t_1_80 = MakeNanos(1, 80);
    NanoTime t_2_50 = MakeNanos(2, 50);
    NanoTime t_2_80 = MakeNanos(2, 80);

    ok &= CheckCmp(&t_1_50, &t_1_50, 0);
    ok &= CheckCmp(&t_1_50, &t_1_80, 1);
    ok &= CheckCmp(&t_1_50, &t_2_50, 1);
    ok &= CheckCmp(&t_1_50, &t_2_80, 1);

    ok &= CheckCmp(&t_1_80, &t_1_80, 0);
    ok &= CheckCmp(&t_1_80, &t_2_50, 1);
    ok &= CheckCmp(&t_1_80, &t_2_80, 1);

    ok &= CheckCmp(&t_2_50, &t_2_50, 0);
    ok &= CheckCmp(&t_2_50, &t_2_80, 1);

    ok &= CheckCmp(&t_2_80, &t_2_80, 0);
    return ok;
}

int main(void)
{
    BOOL ok = TestCompare();

    return ok ? 0 : 1;
}

#endif
