/**
 * @file ebox_device.c Implementation of the device handler of the E-Box (2011) devices for the SOEM master.
 *
 * @todo status, counter, tasain, and control are not handled.
 */

#include "stdincludes.h"
#include "memory.h"
#include "soem_devices.h"

/* EBOX */
#define EBOX_ADC_RESOLUTION     (20.0/200000.0)
#define EBOX_ADC_RANGE          20.0
#define EBOX_ADC_NCHAN          2
#define EBOX_DAC_RESOLUTION     (20.0/65535.0)
#define EBOX_DAC_RANGE          20.0
#define EBOX_DAC_NCHAN          2
#define EBOX_ENC_NCHAN          2
#define EBOX_DIN_NCHAN          8
#define EBOX_DOUT_NCHAN         8
#define EBOX_PWM_NCHAN          2
#define EBOX_PWM_RANGE          2000
#define EBOX_DEV_ID             0x01107001
#define EBOX_DEV_ID2            0x00904004
#define EBOX_DEV_STR            "EBOX"

typedef struct PACKED
{
    uint8_t  status;
    uint8_t  counter;
    uint8_t  din;
    int32_t  ain[2];
    uint32_t tsain;
    int32_t  enc[2]; /* channel 1 becomes timestamp of channel 0 if time stamp mode is enabled over CoE */
} in_Eboxt;

typedef struct PACKED
{
    uint8_t  control;
    uint8_t  dout;
    int16_t  aout[2];
    uint16_t pwmout[2];
} out_Eboxt;

struct EBoxDeviceInstanceStruct {
    DeviceInstance device_instance;

    int64_t enc_position[2]; /**< Cumulative position of both encoders. */
    int32_t enc_last[2];     /**< Last measurement of both encoders. */
};
typedef struct EBoxDeviceInstanceStruct EBoxDeviceInstance;

static void ebox_setup(DeviceInstance *device_instance) {
    EBoxDeviceInstance *ebox_device = (EBoxDeviceInstance *)device_instance;

    ebox_device->enc_position[0] = INT64_MAX;
    ebox_device->enc_position[1] = INT64_MAX;
}

/**
 * Analogue input reading.
 * @param instance Io point instance to compute.
 * @param index Input index number of that io point.
 */
static double read_ain(IoInstance *instance, int index) {
    EBoxDeviceInstance *ebox_di = (EBoxDeviceInstance *)instance->device;
    const in_Eboxt *input_address = ebox_di->device_instance.input_address;

    /* XXX Assumes capture multiplier of 1. */
    return input_address->ain[index] * EBOX_ADC_RESOLUTION;
}

static double ebox_read_ain0(IoInstance *instance) {
    return read_ain(instance, 0);
}
static double ebox_read_ain1(IoInstance *instance) {
    return read_ain(instance, 1);
}

/**
 * Encoder reading
 * @param instance Io point instance to compute.
 * @param index Encoder index number of that io point.
 * @return The new measurement.
 */
static double read_encoder(IoInstance *instance, int index) {
    EBoxDeviceInstance *ebox_di = (EBoxDeviceInstance *)instance->device;
    const in_Eboxt *input_address = ebox_di->device_instance.input_address;

    if (ebox_di->enc_position[0] == INT64_MAX && ebox_di->enc_position[1] == INT64_MAX) { /* First read, setup both encoders. */
        ebox_di->enc_position[0] = 0;
        ebox_di->enc_position[1] = 0;
        ebox_di->enc_last[0] = input_address->enc[0];
        ebox_di->enc_last[1] = input_address->enc[1];
    }

    int32_t movement = input_address->enc[index] - ebox_di->enc_last[index];
    ebox_di->enc_position[index] += movement;
    ebox_di->enc_last[index] = input_address->enc[index];

    /* Only pass the lower 32 bit, to make it fit in a double. */
    return (int32_t)(ebox_di->enc_position[index] & 0xFFFFFFFF);
}

static double ebox_read_enc0(IoInstance *instance) {
    return read_encoder(instance, 0);
}
static double ebox_read_enc1(IoInstance *instance) {
    return read_encoder(instance, 1);
}

/**
 * Read digital input.
 * @param instance Io point of the digital input.
 * @return The read value.
 */
static double ebox_read_din0(IoInstance *instance) {
    DeviceInstance *soem_di = (DeviceInstance *)instance->device;
    const in_Eboxt *input_address = soem_di->input_address;

    return input_address->din;
}

/**
 * Write analogue output.
 * @param instance Io point of the analogue output.
 * @param index Analogue output index number of that io point.
 * @param voltage Value to write to the analgue output.
 */
static void write_aout(IoInstance *instance, int index, double voltage) {
    DeviceInstance *soem_di = (DeviceInstance *)instance->device;
    out_Eboxt *output_address = soem_di->output_address;

    int64_t val = voltage / EBOX_DAC_RESOLUTION;
    int16_t dac_val;
    if (val > 32767) {
        dac_val = 32767;
    } else  if (val < -32767) {
        dac_val = -32767;
    } else {
        dac_val = (int16_t)val;
    }
    output_address->aout[index] = dac_val;
}

static void ebox_write_aout0(IoInstance *instance, double voltage) {
    write_aout(instance, 0, voltage);
}
static void ebox_write_aout1(IoInstance *instance, double voltage) {
    write_aout(instance, 1, voltage);
}

static void ebox_write_dout0(IoInstance *instance, double val) {
    DeviceInstance *soem_di = (DeviceInstance *)instance->device;
    out_Eboxt *output_address = soem_di->output_address;

    output_address->dout = (uint8_t)val;
}

static void write_pwm(IoInstance *instance, double percentage, int index) {
    DeviceInstance *soem_di = (DeviceInstance *)instance->device;
    out_Eboxt *output_address = soem_di->output_address;

    int value = EBOX_PWM_RANGE * percentage / 100.0;
    if (value < 0) {
        value = 0;
    } else if (value > EBOX_PWM_RANGE) {
        value = EBOX_PWM_RANGE;
    }
    output_address->pwmout[index] = value;
}

static void ebox_write_pwm0(IoInstance *instance, double val) {
    write_pwm(instance, val, 0);
}
static void ebox_write_pwm1(IoInstance *instance, double val) {
    write_pwm(instance, val, 1);
}


static const InputIoDefinition ebox_ain0 = {
    {DK_AIN, "ain0", 1}, &ebox_read_ain0
};
static const InputIoDefinition ebox_ain1 = {
    {DK_AIN, "ain1", 1}, &ebox_read_ain1
};

static const InputIoDefinition ebox_din0 = {
    {DK_DIN, "din0", EBOX_DIN_NCHAN}, &ebox_read_din0
};

static const InputIoDefinition ebox_enc0 = {
    {DK_ENC, "enc0", 1}, &ebox_read_enc0
};
static const InputIoDefinition ebox_enc1 = {
    {DK_ENC, "enc1", 1}, &ebox_read_enc1
};

static const OutputIoDefinition ebox_aout0 = {
    {DK_AOUT, "aout0", 1}, &ebox_write_aout0
};
static const OutputIoDefinition ebox_aout1 = {
    {DK_AOUT, "aout1", 1}, &ebox_write_aout1
};

static const OutputIoDefinition ebox_dout0 = {
    {DK_DOUT, "dout0", EBOX_DOUT_NCHAN}, &ebox_write_dout0
};

static const OutputIoDefinition ebox_pwm0 = {
    {DK_PWM, "pwm0", 1}, &ebox_write_pwm0
};
static const OutputIoDefinition ebox_pwm1 = {
    {DK_PWM, "pwm1", 1}, &ebox_write_pwm1
};


static const InputIoDefinition *ebox_inputs[] = {
    &ebox_ain0, &ebox_ain1,
    &ebox_din0,
    &ebox_enc0, &ebox_enc1,
    NULL
};
static const OutputIoDefinition *ebox_outputs[] = {
    &ebox_aout0, &ebox_aout1,
    &ebox_dout0,
    &ebox_pwm0, &ebox_pwm1,
    NULL
};

const DeviceDefinition ebox1_device = {
    EBOX_DEV_STR,
    ebox_inputs, ebox_outputs,
    sizeof(EBoxDeviceInstance),
    ebox_setup,
    EBOX_DEV_ID,
};

const DeviceDefinition ebox2_device = {
    EBOX_DEV_STR,
    ebox_inputs, ebox_outputs,
    sizeof(EBoxDeviceInstance),
    ebox_setup,
    EBOX_DEV_ID2,
};
