/** @file error_storage.c Error storage and retrieval code. */

#include "stdincludes.h"
#include "error_storage.h"

#include <string.h>

/*
/ **
 * Initialize the error storage.
 * @param es Storage to initialize.
 * /
void InitErrorStorage(ErrorStorage *es)
{
    es->error_type = ET_NONE;
    es->message = NULL;
    es->group_index = -1;
    es->sample_index = -1;
}

/ **
 * For any combination of existing error type and new error type, decide whether the
 * new error overrides the existing error.
 * @param existing Existing type of error, currently stored.
 * @param new_type Type of error that may or may not override the existing message.
 * @return Whether the new error message should be stored instead.
 * /
static BOOL ShouldOverwriteExisting(ErrorType existing, ErrorType new_type)
{
    if (new_type == ET_NONE) return FALSE;

    switch (existing) {
        case ET_NONE:
            / * any message overrides no message. * /
            return TRUE;

        case ET_WARNING:
            / * Allow for error and fatal only. * /
            return new_type != ET_WARNING;

        case ET_ERROR:
            / * Allow for fatal only. * /
            return new_type == ET_FATAL;

        case ET_FATAL:
            / * First fatal is preserved. * /
            return FALSE;

        default:
            / * Unknown state, attempt to not crash. * /
            return TRUE;
    }
}

/ **
 * Store a warning in the message if possible. Routine does not overwrite an
 * already existing message.
 * @param es Destination of the warning message.
 * @param message Statically allocated message to store.
 * @param group_index Group number where the error is found (if non-negative).
 * /
void SetWarning(ErrorStorage *es, const char *message, int16_t group_index, int32_t sample_index)
{
    if (!ShouldOverwriteExisting(es->error_type, ET_WARNING)) return;

    es->error_type = ET_WARNING;
    es->message = message;
    es->group_index = group_index;
    es->sample_index = sample_index;
}

/ **
 * Store a non-fatal error in the message if possible. Routine does not overwrite an
 * already existing error or fatal message.
 * @param es Destination of the error message.
 * @param message Statically allocated message to store.
 * @param group_index Group number where the error is found (if non-negative).
 * /
void SetError(ErrorStorage *es, const char *message, int16_t group_index, int32_t sample_index)
{
    if (!ShouldOverwriteExisting(es->error_type, ET_ERROR)) return;

    es->error_type = ET_ERROR;
    es->message = message;
    es->group_index = group_index;
    es->sample_index = sample_index;
}

/ **
 * Store a fatal error in the message if possible. Routine does not overwrite an
 * already existing fatal message.
 * @param es Destination of the error message.
 * @param message Statically allocated message to store.
 * @param group_index Group number where the error is found (if non-negative).
 * /
void SetFatal(ErrorStorage *es, const char *message, int16_t group_index, int32_t sample_index)
{
    if (!ShouldOverwriteExisting(es->error_type, ET_FATAL)) return;

    es->error_type = ET_FATAL;
    es->message = message;
    es->group_index = group_index;
    es->sample_index = sample_index;
}

/ **
 * Copy a stored message.
 * @param dst Destination to write to.
 * @param src Origin to copy from.
 * @return Whethr the error was copied.
 * /
BOOL CopyErrorStorage(ErrorStorage *dst, const ErrorStorage *src)
{
    if (!ShouldOverwriteExisting(dst->error_type, src->error_type)) return FALSE;

    dst->error_type = src->error_type;
    dst->message = src->message;
    dst->group_index = src->group_index;
    dst->sample_index = src->sample_index;
    return TRUE;
}

/ **
 * Retrieve the error message from storage. Should only be called after detecting
 * an error occurred.
 * @param es Error storage to query.
 * @return Statically allocated error message from the storage (do not free() it).
 * /
const char *GetErrorMessage(const ErrorStorage *es)
{
    assert(es->error_type != ET_NONE);
    return es->message;
}
*/

/**
 * Initialize the error info structure.
 * @param err_info Structure to initialize.
 * @param err_text Text of the error.
 * @param err_type Type of the error.
 * @param group_index Index of the signal group, or \c -1.
 */
void InitErrorInfo(ErrorInfo *err_info, const char *err_text, ErrorType err_type, int16_t group_index)
{
    int i;

    err_info->err_text = err_text;
    err_info->err_type = err_type;
    err_info->group_index = group_index;
    err_info->num_occurrences = 0;
    err_info->first_index = -1;
    err_info->last_index = -1;
}

/**
 * Add a point in time when the error occurred.
 * @param err_info Error information of the error that occurred.
 * @param sample_index Sample number to indicate point in time.
 */
void AddSampleIndex(ErrorInfo *err_info, int32_t sample_index)
{
    if (err_info->num_occurrences < UINT16_MAX) err_info->num_occurrences++;
    if (sample_index < 0) return;

    if (err_info->first_index < 0) {
        err_info->first_index = sample_index;
        return;
    }
    err_info->last_index = sample_index;
}

/**
 * Initiaize the collected errors structure.
 * @param collect Collection to initialize.
 */
void InitErrorCollect(ErrorCollect *collect)
{
    int i;

    collect->num_recorded = 0;
    collect->stored_errtypes = 0;
    for (i = 0; i < MAX_ERRORS_RECORDED; i++) {
        InitErrorInfo(&collect->occurred[i], NULL, -1, -1);
    }
}

/**
 * Find an already reported error in the collection.
 * @param collect Collection to search.
 * @param err_text Text of the error.
 * @param err_type Tpye of the error.
 * @param group_index Group index of th error.
 * @return The found error if it exists in the collection or \c NULL
 */
ErrorInfo *FindErrorInfo(ErrorCollect *collect, const char *err_text, ErrorType err_type, int16_t group_index)
{
    size_t i;

    for (i = 0; i < collect->num_recorded; i++) {
        ErrorInfo *info = &collect->occurred[i];
        if (info->err_type != err_type) continue;
        if (info->group_index != group_index) continue;
        if (strcmp(info->err_text, err_text) != 0) continue;
        return info;
    }
    return NULL;
}

/**
 * Find an error to report.
 * @param collect Collection to search.
 * @return The best error to report, or \c NULL if there is no error to report.
 */
ErrorInfo *GetCollectErrorInfo(ErrorCollect *collect)
{
    size_t i;
    ErrorInfo *best_info = NULL;

    for (i = 0; i < collect->num_recorded; i++) {
        ErrorInfo *info = &collect->occurred[i];
        if (info->err_type == ET_FATAL) return info;
        if (best_info == NULL && info->err_type == ET_ERROR) best_info = info;
    }
    return best_info;
}

/**
 * Find or add an error to the collection if possible.
 * @param collect Collection to search or extend.
 * @param err_text Text of the error.
 * @param err_type Tpye of the error.
 * @param group_index Group index of th error.
 * @return Matching or created error that matches with the given values, or \c NULL
 *      if no match existed, and no space was available to add another error.
 */
static ErrorInfo *FindAddErrorInfo(ErrorCollect *collect, const char *err_text, ErrorType err_type, int16_t group_index)
{
    assert(collect != NULL);
    assert(err_text != NULL);

    ErrorInfo *info = FindErrorInfo(collect, err_text, err_type, group_index);
    if (info != NULL) return info;

    if (collect->num_recorded >= MAX_ERRORS_RECORDED) return NULL;

    info = &collect->occurred[collect->num_recorded];
    collect->num_recorded++;
    collect->stored_errtypes |= 1 << err_type;
    InitErrorInfo(info, err_text, err_type, group_index);
    return info;
}

/**
 * Set of add a \a sample_index to an error.
 * @param collect Collection to search or extend.
 * @param err_text Text of the error.
 * @param err_type Tpye of the error.
 * @param group_index Group index of th error.
 * @param sample_index Sample index number that the error occurred.
 */
void AddCollectError(ErrorCollect *collect, const char *err_text, ErrorType err_type, int16_t group_index, int32_t sample_index)
{
    ErrorInfo *info = FindAddErrorInfo(collect, err_text, err_type, group_index);
    if (info != NULL) AddSampleIndex(info, sample_index);
}

