/** @file error_storage.h Data and functions to store error messages while performing data IO. */

#ifndef ERROR_STORAGE_H
#define ERROR_STORAGE_H

/** Types of errors. */
enum ErrorTypeEnum {
    ET_NONE,    /**< Nothing is stored. */
    ET_WARNING, /**< A warning is stored. */
    ET_ERROR,   /**< An error is stored. */
    ET_FATAL,   /**< A fatal error is stored. */
};
typedef enum ErrorTypeEnum ErrorType;

/*
/ ** Storage of a detected error. * /
struct ErrorStorageStruct {
    ErrorType error_type; / **< Stored type of error. * /
    const char *message;  / **< Statically allocated message, or \c NULL. * /
    int16_t group_index;  / **< If non-negative, the group index where the error was found. * /
    int32_t sample_index; / **< Sample index when the error occurred. * /
};
typedef struct ErrorStorageStruct ErrorStorage;

void InitErrorStorage(ErrorStorage *es);
void SetWarning(ErrorStorage *es, const char *message, int16_t group_index, int32_t sample_index);
void SetError(ErrorStorage *es, const char *message, int16_t group_index, int32_t sample_index);
void SetFatal(ErrorStorage *es, const char *message, int16_t group_index, int32_t sample_index);
BOOL CopyErrorStorage(ErrorStorage *dst, const ErrorStorage *src);

/ **
 * Get the type of error stored in the storage.
 * @param es Error storage to query.
 * @return Type of the stored error.
 * /
#define GetStoredErrorType(es) ((es)->error_type)

/ **
 * Test whether the storage has a message at all.
 * @param es Error storage to test.
 * @return Whether the provided storage has a message.
 * /
#define HasStoredMessage(es) (GetStoredErrorType(es) != SET_NONE)
*/

/** Number of sample indices recorded of each error. */
#define NUM_INDICES_RECORDED 4

/** Information about a single error that occurred. */
struct ErrorInfoStruct {
    const char *err_text; /**< Text of the error, \c NULL if not initialized. */
    ErrorType err_type;   /**< Type of error. */
    int16_t group_index;  /**< If non-negative, the group index where the error was found. */
    uint16_t num_occurrences; /*< Number of occurrances of this error. */
    int32_t first_index;  /**< First non-negative sample index of this error. */
    int32_t last_index;   /**< Last non-negative sample index of this error. */
};
typedef struct ErrorInfoStruct ErrorInfo;

void InitErrorInfo(ErrorInfo *err_info, const char *err_text, ErrorType err_type, int16_t group_index);
void AddSampleIndex(ErrorInfo *err_info, int32_t sample_index);

#define GetErrorText(err_info) ((err_info)->err_text)

const char *GetErrorSeverity(const ErrorInfo *err_info);

/** Number of different errors that are recorded. */
#define MAX_ERRORS_RECORDED 16

/**
 * Collected errors that occurred.
 * @todo For each type of error, record whether that type of error was added.
 */
struct ErrorCollectStruct {
    uint32_t num_recorded; /**< Number of errors collected in the structure. */
    uint8_t stored_errtypes; /**< Bitmask of stored error types in the collection. */
    ErrorInfo occurred[MAX_ERRORS_RECORDED]; /**< Details of each recorded error. */
};
typedef struct ErrorCollectStruct ErrorCollect;

void InitErrorCollect(ErrorCollect *collect);
ErrorInfo *FindErrorInfo(ErrorCollect *collect, const char *err_text, ErrorType err_type, int16_t group_index);

/**
 * Does the error collection have a fatal error?
 * @param collect Collected errors.
 * @return Whether a fatal error exists in the collection.
 */
#define CollectHasFatalError(collect) ((((collect)->stored_errtypes) & (1 << ET_FATAL)) != 0)


/**
 * Does the error collection have an error or a fatal error?
 * @param collect Collected errors.
 * @return Whether an error or a fatal error exists in the collection.
 */
#define CollectHasError(collect) ((((collect)->stored_errtypes) & ((1 << ET_FATAL) | (1 << ET_ERROR))) != 0)

ErrorInfo *GetCollectErrorInfo(ErrorCollect *collect);

/**
 * Get the error message of the best error to report.
 * @param collect Collected errors.
 * @return Error message text of the best error to report.
 * @pre An error must exist, that is #CollectHasError or #CollectHasFatalError must hold.
 */
#define GetCollectErrorMessage(collect) (GetCollectErrorInfo(collect)->err_text)

void AddCollectError(ErrorCollect *collect, const char *err_text, ErrorType err_type, int16_t group_index, int32_t sample_index);

#endif
