/** @file msgbox.c Control messages between matlab thread and real-time thread. */

#include "stdincludes.h"

#include "msgbox.h"

/**
 * Initialize the provided message box with initial values.
 * @param mbox Message box to initialize.
 */
void MessageBoxInit(MessageBox *mbox)
{
    pthread_mutex_init(&mbox->lock, NULL);
    pthread_cond_init(&mbox->parent_block, NULL);

    mbox->parent_mask = 0;;
    mbox->parent_action = PAR_NONE;
    mbox->realtime_state = RTS_NOT_RUNNING;
}

/**
 * Destroy the message box.
 * @param mbox Message box to destroy.
 * @note No real-time thread may be running at the time of the call.
 */
void MessageBoxDestroy(MessageBox *mbox)
{
    pthread_mutex_destroy(&mbox->lock);
    pthread_cond_destroy(&mbox->parent_block);
}

/**
 * Start a new realtime thread.
 * @param mbox Message box shared with the real-time thread.
 * @param thread Thread function to start.
 * @return Whether creation was successful.
 */
BOOL ParentStartRealtime(MessageBox *mbox, void *(*thread)(void *data))
{
    int errcode;
    extern void *RealTimeThread(void *arg);

    pthread_mutex_lock(&mbox->lock);
    assert(mbox->realtime_state == RTS_NOT_RUNNING);

    errcode = pthread_create(&mbox->realtime_id, NULL, thread, mbox);
    if (errcode == 0) mbox->realtime_state = RTS_RUNNING;

    pthread_mutex_unlock(&mbox->lock);
    return errcode == 0;
}

/**
 * Parent waits for at least one buffer becoming available in parent stage from the set of
 * indicated ring buffers.
 * @param mbox Message box to wait on.
 * @param mask Set of ring buffers being blocked on due to lack of buffers in parent stage.
 * @return #PRS_NOTHING, or #PRS_STOP_REALTIME if real-time thread has exited.
 */
PollResult ParentWaitMask(MessageBox *mbox, uint64_t mask)
{
    PollResult pr = PRS_NOTHING;

    pthread_mutex_lock(&mbox->lock);
    mbox->parent_mask = mask;
    do {
        if (mbox->realtime_state == RTS_EXITING || mbox->realtime_state == RTS_NOT_RUNNING) {
            pr = PRS_STOP_REALTIME;
            break;
        }
        mbox->parent_action = PAR_WAIT_NOTIFY;
        pthread_cond_wait(&mbox->parent_block, &mbox->lock);
    } while(mbox->parent_mask == mask);

    mbox->parent_action = PAR_NONE;
    mbox->parent_mask = 0;
    pthread_mutex_unlock(&mbox->lock);

    return pr;
}

/**
 * Parent stops the real-time thread.
 * @param mbox Message box used for communication.
 */
void ParentStopRealtime(MessageBox *mbox)
{
    pthread_mutex_lock(&mbox->lock);
    if (mbox->realtime_state == RTS_NOT_RUNNING) {
        /* Real-time process is not running, nothing needs to be done. */
        pthread_mutex_unlock(&mbox->lock);
        return;
    }

    while (mbox->realtime_state == RTS_RUNNING) {
        mbox->parent_action = PAR_WAIT_DIE;
        pthread_cond_wait(&mbox->parent_block, &mbox->lock);
    }
    pthread_mutex_unlock(&mbox->lock);

    pthread_join(mbox->realtime_id, NULL);
    mbox->realtime_state = RTS_NOT_RUNNING;
    /* Done. */
}

/**
 * Parent polls the message box for new messages.
 * @param mbox Message box used for communication.
 * @return Polling result. If it is #PRS_STOP_REALTIME, an error may have been
 *      written as well.
 */
PollResult ParentPollBox(MessageBox *mbox)
{
    PollResult pr = PRS_NOTHING;

    pthread_mutex_lock(&mbox->lock);

    if (mbox->realtime_state == RTS_EXITING) pr = PRS_STOP_REALTIME;

    pthread_mutex_unlock(&mbox->lock);
    return pr;
}

BOOL RingsCheckMask(uint64_t mask); /* Generated function doing the actual checking. */

/**
 * Real-time thread checks for new message in the message box.
 * @param mbox Message box used for communication.
 * @return #PRS_NOTHING, or #PRS_STOP_REALTIME if the parent thread wants to
 *      stop data acquisition.
 */
PollResult RealTimePoll(MessageBox *mbox)
{
    PollResult pr;

    pthread_mutex_lock(&mbox->lock);

    /* Check whether parent has messages for realtime thread. */
    switch (mbox->parent_action) {
        case PAR_NONE:
            pr = PRS_NOTHING;
            break;

        case PAR_WAIT_NOTIFY:
            if (RingsCheckMask(mbox->parent_mask)) {
                mbox->parent_mask = 0; /* Let parent know a buffer was detected. */
                pthread_cond_signal(&mbox->parent_block);
            }
            pr = PRS_NOTHING;
            break;

        case PAR_WAIT_DIE:
            pr = PRS_STOP_REALTIME;
            /* Do not wake parent thread, wait until real-time exits. */
            break;

        default:
            NOT_REACHED();
    }
    pthread_mutex_unlock(&mbox->lock);
    return pr;
}

/**
 * Real-time thread stops executing.
 * @param mbox Message box used for communication.
 */
void RealTimeExit(MessageBox *mbox)
{
    pthread_mutex_lock(&mbox->lock);
    mbox->realtime_state = RTS_EXITING;

    if (mbox->parent_action == PAR_WAIT_DIE || mbox->parent_action == PAR_WAIT_NOTIFY) {
        pthread_cond_signal(&mbox->parent_block);
    }
    pthread_mutex_unlock(&mbox->lock);

    pthread_exit(NULL);
    /* Never returns, that's all folks, thanks for watching! */
}
