/** msgbox.h HEader file of the message box. */

#ifndef MSGBOX_H
#define MSGBOX_H

#include <pthread.h>

/** Actions of the parent that involve the real-time thread. */
enum ParentActionEnum {
    PAR_NONE,        /**< Parent does not need interaction with real-time thread. */
    PAR_WAIT_NOTIFY, /**< Parent is blocked on getting a notification of a new marker. */
    PAR_WAIT_DIE,    /**< Parent waits for death of real-time thread. */
};
typedef enum ParentActionEnum ParentAction;

/** States of the real-time thread. */
enum RealTimeStateEnum {
    RTS_NOT_RUNNING, /**< Real-time thread is not running. */
    RTS_RUNNING,     /**< Real-time thread is running. */
    RTS_EXITING,     /**< Real-time thread is exiting (or has exited). */
};
typedef enum RealTimeStateEnum RealTimeState;

enum PollResultEnum {
    PRS_NOTHING,
    PRS_STOP_REALTIME,
    PRS_ERROR_FOUND,
};
typedef enum PollResultEnum PollResult;

/** Data structure for messaging and control of the real-time thread from the parent matlab thread. */
struct MessageBoxStruct {
    pthread_t realtime_id;        /**< Process thread ID of the realtime thread. */
    pthread_mutex_t lock;         /**< Access lock for the message box. */
    pthread_cond_t parent_block;  /**< Condition variable for waiting by the parent. */

    uint64_t parent_mask;         /**< Nask of ring buffer groups waiting for a buffer in parent stage. */
    ParentAction parent_action;   /**< Current action of the parent. */
    RealTimeState realtime_state; /**< State of the real-time thread. */
};
typedef struct MessageBoxStruct MessageBox;

void MessageBoxInit(MessageBox *mbox);
void MessageBoxDestroy(MessageBox *mbox);

/* Parent message box functions. */
BOOL ParentStartRealtime(MessageBox *mbox, void *(*thread)(void *data));
PollResult ParentWaitMask(MessageBox *mbox, uint64_t mask);
void ParentStopRealtime(MessageBox *mbox);
PollResult ParentPollBox(MessageBox *mbox);

/* Realtime message box functions. */
PollResult RealTimePoll(MessageBox *mbox);
void RealTimeExit(MessageBox *mbox);

#endif
