/** @file nanotime.c Nano-second time computing. */

#include "stdincludes.h"
#include "nanotime.h"

#ifdef USE_GETTIMEOFDAY
#include <sys/time.h>
#endif

/**
 * Construct a nano-time from a seconds and nano seconds component.
 * @param secs Seconds since Epoch.
 * @param nanos Nanoseconds expressing the fraction of the last second.
 * @return The created nano time.
 */
NanoTime MakeNanos(uint64_t secs, uint32_t nanos)
{
    NanoTime ntime;

    assert(nanos < ONE_SEC);
    ntime.tv_sec = secs;
    ntime.tv_nsec = nanos;
    return ntime;
}

/**
 * Construct a nano-time from a single 64 bit integer.
 * @param nsecs Time as 64 integer, in nano-seconds.
 * @return The created nano time.
 */
NanoTime MakeNano64(uint64_t nsecs)
{
    uint32_t nanos;

    if (nsecs < ONE_SEC) return MakeNanos(0, nsecs);

    nanos = nsecs % ONE_SEC;
    return MakeNanos(nsecs / ONE_SEC, nanos);
}

/**
 * Construct a floating point value for the provided time, for easier debugging.
 * @param t Time to convert.
 * @return The converted time.
 */
double ConvertNano(const NanoTime *t)
{
    double fp = ((double)t->tv_nsec) / ONE_SEC;
    fp += t->tv_sec;
    return fp;
}

/**
 * Get current time as nano time.
 * @return Current nano time, up to micro second.
 */
NanoTime NowNanos(void)
{
    NanoTime ntime;

#ifdef USE_GETTIMEOFDAY
    struct timeval tv;

    if (gettimeofday(&tv, NULL) != 0) {
        ntime.tv_sec = 0;
        ntime.tv_nsec = 0;
        return ntime;
    }
    ntime.tv_sec = tv.tv_sec;
    ntime.tv_nsec = tv.tv_usec * 1000;
#else
    if (clock_gettime(CLOCK_REALTIME, &ntime) != 0) {
        ntime.tv_sec = 0;
        ntime.tv_nsec = 0;
    }
#endif
    return ntime;
}

/**
 * Sum two nano times.
 * @param nt1 First nano time to add.
 * @param nt2 Second nano time to add.
 * @return Sum of both times.
 */
NanoTime AddNanos(const NanoTime *nt1, const NanoTime *nt2)
{
    NanoTime ntime;
    ntime.tv_sec = nt1->tv_sec + nt2->tv_sec;

    assert(nt1->tv_nsec < ONE_SEC);
    assert(nt2->tv_nsec < ONE_SEC);

    uint32_t nsum = nt1->tv_nsec + nt2->tv_nsec;
    if (nsum > ONE_SEC) {
        ntime.tv_sec++;
        ntime.tv_nsec = nsum - ONE_SEC;
    } else {
        ntime.tv_nsec = nsum;
    }
    return ntime;
}

/**
 * Increase a nano time with another nano time.
 * @param [inout] nt Nano-time to update.
 * @param [in] add Nano-time to add.
 */
void IncreaseNano(NanoTime *nt, const NanoTime *add)
{
    nt->tv_sec += add->tv_sec;

    assert(nt->tv_nsec < ONE_SEC);
    assert(add->tv_nsec < ONE_SEC);

    uint32_t nsum = nt->tv_nsec + add->tv_nsec;
    if (nsum > ONE_SEC) {
        nt->tv_sec++;
        nt->tv_nsec = nsum - ONE_SEC;
    } else {
        nt->tv_nsec = nsum;
    }
}

/**
 * Compare left nano time with right nano time.
 * @param ln Left nano time to compare.
 * @param rn Right nano time to compare.
 * @return \c 0 if both nano times are equal, positive if \a ln is after \a rn,
 *      negative if \a rn is after \a ln.
 */
int CmpNanos(const NanoTime *ln, const NanoTime *rn)
{
    if (ln->tv_sec < rn->tv_sec) return 1;
    if (ln->tv_sec > rn->tv_sec) return -1;

    if (ln->tv_nsec < rn->tv_nsec) return 1;
    if (ln->tv_nsec > rn->tv_nsec) return -1;
    return 0;
}

/**
 * Copy nano time.
 * @param [out] dst Destination nano time.
 * @param src Source nano time.
 */
void CopyNanos(NanoTime *dst, const NanoTime *src)
{
    dst->tv_sec = src->tv_sec;
    dst->tv_nsec = src->tv_nsec;
}

