/** @file soem_devices.c Implementatie van soem device and io point access. */

#include "ethercattype.h"
#include "nicdrv.h"
#include "ethercatmain.h"

#include "stdincludes.h"
#include "soem_devices.h"
#include "memory.h"

#include <string.h>


extern DeviceDefinition ek1100_device;
extern DeviceDefinition el1004_device, el1008_device, el1014_device, el1018_device, el1819_device;
extern DeviceDefinition el2004_device, el2008_device, el2809_device;
extern DeviceDefinition el3102_device, el3104_device;
extern DeviceDefinition el4038_device, el4132_device, el4134_device;
extern DeviceDefinition el5002_device, el5101_device, el5152_device;
extern DeviceDefinition ebox1_device, ebox2_device;


/** Known EtherCaT devices, \c NULL terminated. */
static const DeviceDefinition *devices[] = {
    &ek1100_device,
    &el1004_device, &el1008_device, &el1014_device, &el1018_device, &el1819_device,
    &el2004_device, &el2008_device, &el2809_device,
    &el3102_device, &el3104_device,
    &el4038_device, &el4132_device, &el4134_device,
    &el5002_device, &el5101_device, &el5152_device,
    &ebox1_device, &ebox2_device,
    NULL
};

/* {{{ static const DeviceDefinition *FindDevice(uint32_t dev_id) */
/**
 * Find the device definition with the provided \a dev_id.
 * @param dev_id Device identification to find.
 * @return Matching device definition if it can be found, else \c NULL.
 */
static const DeviceDefinition *FindDevice(uint32_t dev_id)
{
    const DeviceDefinition **dd = devices;
    while (*dd != NULL) {
        if ((*dd)->device_id == dev_id) return *dd;
        dd++;
    }
    return NULL;
}
/* }}} */

/* {{{ static size_t CountRecognizedSlaves(void) */
/**
 * Count the number of slaves recognized by the library handling code.
 * @return Number of recognized slaves.
 */
static size_t CountRecognizedSlaves(void)
{
    size_t index;
    size_t count = 0;

    for (index = 1; index <= ec_slavecount; index++) {
        if (FindDevice(ec_slave[index].eep_id) != NULL) count++;
    }
    return count;
}
/* }}} */
/* {{{ static DeviceInstance **AllocateDeviceInstances(Memory *mem, const size_t count) */
/**
 * Allocate the device instances, and initialize it with \c NULL.
 * @param mem Memory manager.
 * @param count Number of recognized slaves.
 * @return Initialized allocated memory for an array of pointers to device instances, or \c NULL.
 * @note Should have at least one recongized slave.
 */
static DeviceInstance **AllocateDeviceInstances(Memory *mem, const size_t count)
{
    size_t index;

    assert(count > 0);
    DeviceInstance **dev_instances = MemoryAllocate(mem, sizeof(DeviceInstance *) * count);
    if (dev_instances == NULL) return NULL;

    /* Initialize all pointers to NULL. */
    for (index = 0; index < count; index++) dev_instances[index] = NULL;

    return dev_instances;
}
/* }}} */
/* {{{ static BOOL SetupDeviceInstances(DeviceInstance **dev_instances, Memory *mem, const size_t inst_count) */
/**
 * Create device instances for the recognized slaves, and initialize them.
 * @param dev_instances Pointer array to fill.
 * @param mem Memory manager to use for requesting memory.
 * @param inst_count Number of recongized slaves (equal to length of \a dev_instances).
 * @return Whether device instance creation was successful.
 */
static BOOL SetupDeviceInstances(DeviceInstance **dev_instances, Memory *mem, const size_t inst_count)
{
    size_t index, index2, init_count;
    size_t msize;
    DeviceInstance *di, *di2;
    const DeviceDefinition *dd;

    /* Allocation and basic initialization of the device instances. */
    init_count = 0;
    assert(inst_count > init_count);
    for (index = 1; index <= ec_slavecount; index++) {
        dd = FindDevice(ec_slave[index].eep_id);
        if (dd == NULL) continue; /* Skip unrecognized slaves. */

        /* Obtain memory for the device instance. */
        msize = dd->malloc_size;
        assert(msize >= sizeof(DeviceInstance));
        assert(msize <= sizeof(DeviceInstance) + 24); /* Arbitrary upper limit, increase if needed. */
        di = MemoryAllocate(mem, msize);
        if (di == NULL) return FALSE;
        di->device_definition = dd;
        di->instance_number = UINT16_MAX; /* Is changed below. */
        di->input_address = ec_slave[index].inputs;
        di->output_address = ec_slave[index].outputs;
        /* Custom device instance setup done later. */

        dev_instances[init_count] = di;
        init_count++;
        if (init_count == inst_count) break;
    }
    assert(init_count == inst_count); /* Else we lost some recognized slave! */

    /* Setup device counts, we use the textual name of recognized devices, as that is
     * closer to the user perception of "the same device".
     */
    for (index = 0; index < inst_count; index++) {
        di = dev_instances[index];
        if (di->instance_number != UINT16_MAX) continue;

        di->instance_number = 0;
        init_count = 1; /* Free number counter. */
        for (index2 = index + 1; index2 < inst_count; index2++) {
            di2 = dev_instances[index2];
            if (di2->instance_number != UINT16_MAX) continue;
            if (strcmp(di->device_definition->name, di2->device_definition->name) == 0) {
                di2->instance_number = init_count;
                init_count++;
            }
        }
    }

    /* Do custom setup for the device instances that have it. */
    for (index = 0; index < inst_count; index++) {
        di = dev_instances[index];
        dd = di->device_definition;
        if (dd->setup != NULL) (*dd->setup)(di);
    }
    return TRUE;
}
/* }}} */

/**
 * Instantiate recognized slaves.
 * @param mem Memory manager to use for requesting memory.
 * @param [out] device_count Number of recognized devices. Non-zero count while the function
 *      returns \c NULL indicates a memory or initialization error. Zero-count indicates no
 *      slave is recognized at all.
 * @return Table with pointer to instantiated device, or \c NULL if an error occurred.
 */
DeviceInstance **InstantiateDevices(Memory *mem, size_t *device_count)
{
    size_t count;
    DeviceInstance **dev_instances;

    /* Count devices. */
    count = CountRecognizedSlaves();
    *device_count = count;
    if (count == 0) return NULL;

    dev_instances = AllocateDeviceInstances(mem, count);
    if (dev_instances == NULL) return NULL;

    if (!SetupDeviceInstances(dev_instances, mem, count)) return NULL;
    return dev_instances;
}

/**
 * Construct io instances of the found devices.
 * @param dev_instances Found device instances.
 * @param dev_count Number of found devices in \a dev_instances.
 * @param mem Memory manager to use for requesting memory.
 * @param [out] num_iis Number of io instances return to the caller.
 * @return The collection io points of the instantiated devices, or \c NULL if not enough memory.
 */
IoInstance *InstantiateIoPoints(DeviceInstance **dev_instances, size_t dev_count, Memory *mem, size_t *num_iis)
{
    IoInstance *iis, *ii;
    const DeviceDefinition *dd;
    size_t iis_count, iis_count2;
    size_t index, io_index;

    iis_count = 0;
    for (index = 0; index < dev_count; index++) {
        dd = dev_instances[index]->device_definition;
        for (io_index = 0; ; io_index++) {
            const InputIoDefinition *in_def = dd->input_io_definitions[io_index];
            if (in_def == NULL) break;
            iis_count++;
        }
        for (io_index = 0; ; io_index++) {
            const OutputIoDefinition *out_def = dd->output_io_definitions[io_index];
            if (out_def == NULL) break;
            iis_count++;
        }
    }

    *num_iis = iis_count;
    if (iis_count == 0) return NULL;

    iis = MemoryAllocate(mem, sizeof(IoInstance) * iis_count);
    if (iis == NULL) return NULL;

    /* Add inputs. */
    iis_count2 = 0;
    for (index = 0; index < dev_count; index++) {
        dd = dev_instances[index]->device_definition;
        for (io_index = 0; iis_count2 < iis_count; io_index++) {
            const InputIoDefinition *in_def = dd->input_io_definitions[io_index];
            if (in_def == NULL) break;

            ii = &iis[iis_count2];
            ii->device = dev_instances[index];
            ii->io_definition = &in_def->io_definition; /* double (indef->read)(IoInstance *) also exists. */
            iis_count2++;
        }
    }
    /* Add outputs. */
    for (index = 0; index < dev_count; index++) {
        dd = dev_instances[index]->device_definition;
        for (io_index = 0; iis_count2 < iis_count; io_index++) {
            const OutputIoDefinition *out_def = dd->output_io_definitions[io_index];
            if (out_def == NULL) break;

            ii = &iis[iis_count2];
            ii->device = dev_instances[index];
            ii->io_definition = &out_def->io_definition; /* void (outdef->write)(IoInstance *, double) also exists. */
            iis_count2++;
        }
    }

    /* Only return the memory if it is initialized completely! */
    if (iis_count2 == iis_count) return iis;
    return NULL;
}

/**
 * Find the io instance that belongs to the io point of the requested device type and number.
 * @param iis Io instances.
 * @param num_iis Number of io instances in \a iis.
 * @param dev_name Name of the device to search.
 * @param dev_index Index number of the device to search.
 * @param io_point_name Name of the io instance.
 * @return The io instance if the requested entry can be found, or \c NULL.
 */
IoInstance *FindIoInstance(IoInstance *iis, size_t num_iis, const char *dev_name, size_t dev_index, const char *io_point_name)
{
    size_t index;

    for (index = 0; index < num_iis; index++) {
        IoInstance *ii = &iis[index];
        DeviceInstance *di = ii->device;
        if (di->instance_number != dev_index) continue;
        if (strcmp(di->device_definition->name, dev_name) != 0) continue;
        if (strcmp(ii->io_definition->name, io_point_name) == 0) return ii;
    }
    return NULL;
}

/**
 * Find the \a kind_index-'th io instance that matches the given \a kind.
 * @param iis Io instances.
 * @param num_iis Number of io instances in \a iis.
 * @param kind Io instance kind.
 * @param kind_index Index number of the given kind to look for.
 * @return The io instance if the requested entry can be found, or \c NULL.
 */
IoInstance *FindIoKind(IoInstance *iis, size_t num_iis, IoKind kind, size_t kind_index)
{
    size_t index;

    for (index = 0; index < num_iis; index++) {
        IoInstance *ii = &iis[index];
        if (ii->io_definition->kind != kind) continue;
        if (kind_index == 0) return ii;
        kind_index--;
    }
    return NULL;
}
