/** @file soem_handling.c Code for interfacing Soem. */

#include "ethercattype.h"
#include "nicdrv.h"
#include "ethercatmain.h"
#include "ethercatconfig.h"
#include "ethercatdc.h"

#include "stdincludes.h"
#include "memory.h"

#include "soem_handling.h"

#include <unistd.h>
#include <sys/types.h>

/**
 * Initialize the Soem state data structure.
 * @param ssd Soem state data to initialize.
 */
void InitializeSoemState(SoemStateData *ssd)
{
    ssd->state = SSE_UNCONNECTED;
    ssd->dev_instances = NULL;
    ssd->io_points = NULL;
    ssd->num_device_instances = 0;
    ssd->num_io_points = 0;
    ssd->error = NULL;
}


/**
 * Wake Soem, and let it find connected slaves.
 * @param ssd Current state of Soem.
 * @param mem Memory manager.
 * @param nics List of ethernet ports to try to connect to
 * @return Whether wakeup succeeded.
 * @note Soem should be in non-active state when calling the function.
 */
BOOL SoemEngineWakeup(SoemStateData *ssd, Memory *mem, const char *nics[])
{
    void *mem_start;
    size_t mem_size, mem_needed;

    assert(ssd->state == SSE_UNCONNECTED);

    /* 1. Make a connection. */
    while (*nics != NULL) {
        if (ec_init(*nics) > 0) break;
        nics++;
    }

    if (*nics == NULL) {
        /* No connectable NIC found, no EtherCaT slaves, not allowed access, or something else. */
        if (geteuid() == 0) {
            ssd->error = "Cannot connect to any NIC (Correct port? EtherCaT slaves connected?)";
        } else {
            ssd->error = "Cannot connect to any NIC (Correct port? EtherCaT slaves connected?  Root access for raw socket?)";
        }

        ShutdownSoem(ssd);
        return FALSE;
    }

    ssd->state = SSE_CONNECTED; /* We have a connection! */

    /* 2. Find and auto-config slaves. */
    if (ec_config_init(FALSE) <= 0) {
        ssd->error = "No EtherCaT devices found";
        ShutdownSoem(ssd);
        return FALSE;
    }

    /* 3. Setup memory. */
    mem_start = MemoryAllocateAll(mem, &mem_size);
    if (mem_start == NULL) {
        ssd->error = "No memory available";
        ShutdownSoem(ssd);
        return FALSE;
    }
    mem_needed = ec_config_map(mem_start);
    if (mem_needed > mem_size) {
        ssd->error = "Not enough memory for EtherCaT IO-map";
        ShutdownSoem(ssd);
        return FALSE;
    }
    MemoryAllocateSetNeeded(mem, mem_start, mem_needed);

    ec_configdc(); /* Configure distributed clock (if there exist slaves that support it). */
    ec_statecheck(0, EC_STATE_SAFE_OP, EC_TIMEOUTSTATE); /* Verify IO of the call. */

    ssd->state = SSE_WOKEN;
    ssd->error = NULL;
    return TRUE;
}

/**
 * Switch to the given EtherCaT state.
 * @param new_state State to switch to.
 * @note Function may fail to switch the slaves to the desired state. Check master state afterwards.
 */
static void SwitchSoemState(ec_state new_state)
{
    ec_slave[0].state = new_state;
    ec_writestate(0); /* Broadcast new state. */
    ec_statecheck(0, new_state, EC_TIMEOUTSTATE); /* Wait for the new state (or give up). */
}

/**
 * Start running.
 * @param ssd Soem state data.
 * @return Whether starting succeeded.
 * @note Soem engine should be awake when calling this.
 */
BOOL SoemEngineStart(SoemStateData *ssd)
{
    /* Perform one process data cycle to init SM(?) in all slaves. */
    ec_send_processdata(); /* XXX Verify IO of the call. */
    ec_receive_processdata(EC_TIMEOUTRET); /* XXX Verify IO of the call. */

    /* Request operational state. */
    SwitchSoemState(EC_STATE_OPERATIONAL);
    if (ec_slave[0].state != EC_STATE_OPERATIONAL) {
        ssd->error = "Not all EtherCaT devices switched to Operational state";
        return FALSE;
    }

    ssd->state = SSE_RUNNING;
    ssd->error = NULL;
    return TRUE;
}

/**
 * Detect slave devices, and initialize all io points.
 * @param ssd Soem state data.
 * @param mem Memory manager.
 * @return Setup succeeded.
 * @pre Soem needs to be awake but not running.
 */
BOOL SoemSetupIoPoints(SoemStateData *ssd, Memory *mem)
{
    assert(ssd->state == SSE_WOKEN);

    /* Detect EtherCaT devices. */
    ssd->dev_instances = InstantiateDevices(mem, &ssd->num_device_instances);
    if (ssd->dev_instances == NULL) {
        ssd->error = (ssd->num_device_instances == 0) ? "No devices detected" : "Devices initialization error";
        ssd->num_device_instances = 0;
        return FALSE;
    }

    /* Construct available io points. */
    ssd->io_points = InstantiateIoPoints(ssd->dev_instances, ssd->num_device_instances, mem, &ssd->num_io_points);
    if (ssd->io_points == NULL) {
        ssd->error = "No room to allocate io points";
        return FALSE;
    }

    ssd->error = NULL;
    return TRUE;
}

/**
 * Perform an EtherCaT data IO cycle, and report about correctness of the operation.
 * @param ssd Soem state data.
 * @return Whether IO succeeded (all slaves behave as expected).
 */
BOOL SoemDataIo(SoemStateData *ssd)
{
    int real_wkc, expected_wkc;

    ec_send_processdata();
    real_wkc = ec_receive_processdata(EC_TIMEOUTRET);
    expected_wkc = ec_group[0].outputsWKC * 2 + ec_group[0].inputsWKC;

    if (real_wkc == expected_wkc) {
        ssd->error = NULL;
        return TRUE;
    }
    ssd->error = "Slave work-counters fail to match (data-sync error?)";
    return FALSE;
}

/**
 * Tear down Soem to non-active state.
 * @param ssd Current state of Soem.
 */
void ShutdownSoem(SoemStateData *ssd)
{
    if (ssd->state == SSE_WOKEN || ssd->state == SSE_RUNNING) {
        SwitchSoemState(EC_STATE_SAFE_OP);
        SwitchSoemState(EC_STATE_PRE_OP);
        /* State switches may fail, but we're stopping anyway. */
        ssd->state = SSE_CONNECTED;
    }

    if (ssd->state == SSE_CONNECTED) {
        ec_close();
        ssd->state = SSE_UNCONNECTED;
    }

    /* SSE_UNCONNECTED, nothing needs to be done. */

    /* Make the io devices and instances unreachable. Note that the memory manager
     * doesn't require explicit release of memory.
     */
    ssd->io_points = NULL;
    ssd->num_io_points = 0;
    ssd->dev_instances = NULL;
    ssd->num_device_instances = 0;
}
