/** @file soem_handling.h Declarations for interfacing Soem. */

#ifndef SOEM_HANDLING_H
#define SOEM_HANDLING_H

#include "soem_devices.h"

typedef struct MemoryStruct Memory;

/** States of the Soem library (for some relevant sub-set) */
enum SoemStateEnum {
    SSE_UNCONNECTED, /**< Soem library is not active at all. */
    SSE_CONNECTED,   /**< Soem library is connected to a port. */
    SSE_WOKEN,       /**< Soem library woke up successfully. */
    SSE_RUNNING,     /**< Soem library is running. */
};
typedef enum SoemStateEnum SoemState;

/** Data storage for Soem-related variables. */
struct SoemStateDataStruct {
    SoemState state;   /**< State of the Soem library. */
    DeviceInstance **dev_instances; /**< Array of pointers to found devices. */
    IoInstance *io_points; /**< Array of io points of all recognized devices. */
    size_t num_device_instances; /**< Number of detected devices in #dev_instances. */
    size_t num_io_points; /**< Number of found #io_points. */
    const char *error; /**< Description of the error of the last call, if any. */
};
typedef struct SoemStateDataStruct SoemStateData;

void InitializeSoemState(SoemStateData *ssd);

/**
 * Get the newest error message from the Soem state. Typically only useful immediately
 * after a failed call. Returned message is statically allocated, do not free it.
 * @param ssd Soem state data.
 * @return Pointer to error message if one exists, else \c NULL.
 */
#define GetSoemStateError(ssd) ((ssd)->error)

BOOL SoemEngineWakeup(SoemStateData *ssd, Memory *mem, const char *nics[]);
BOOL SoemSetupIoPoints(SoemStateData *ssd, Memory *mem);
BOOL SoemEngineStart(SoemStateData *ssd);
BOOL SoemDataIo(SoemStateData *ssd);
void ShutdownSoem(SoemStateData *ssd);

/**
 * Find the io instance that belongs to the io point of the requested device type and number.
 * @param ssd Soem state data.
 * @param dev_name Name of the device to search.
 * @param dev_index Index number of the device to search.
 * @param io_point_name Name of the io instance.
 * @return The io instance if the requested entry can be found, or \c NULL.
 */
#define SoemFindIoInstance(ssd, dev_name, dev_index, io_point_name) \
        FindIoInstance((ssd)->io_points, (ssd)->num_io_points, (dev_name), (dev_index), (io_point_name))

/**
 * Find the \a kind_index-'th io instance that matches the given \a kind.
 * @param ssd Soem state data.
 * @param kind Io instance kind.
 * @param kind_index Index number of the given kind to look for.
 * @return The io instance if the requested entry can be found, or \c NULL.
 */
#define SoemFindIoKind(ssd, kind, kind_index) \
        FindIoKind((ssd)->io_points, (ssd)->num_io_points, (kind), (kind_index))

#endif
