/** @file test_buffer.c Test code for testing the buffer implementation. */

/* gcc -o test_buffer -Wall -g -DTEST_BUFFER -Isrc src/test_buffer.c src/buffer.c */
#ifdef TEST_BUFFER

#include "stdincludes.h"
#include "buffer.h"

static void SetData(double *base, int length, double value, double increment)
{
    while (length > 0) {
        *base = value;
        base++;
        length--;
        value += increment;
    }
}

static BOOL CheckData(const double *base, int length, double value, double increment)
{
    while (length > 0) {
        if (*base != value) return FALSE;
        base++;
        length--;
        value += increment;
    }
    return TRUE;
}

static void PrintResult(const char *msg, BOOL val)
{
    printf("%s: %s\n", msg, val ? "ok" : "                                            ERROR");
}

static void TestBufferFill(void)
{
    double data[50];
    double values[50];
    Buffer b;
    int i, u, j;

    SetData(data, LENGTH_OF(data), 0, 0);

    /* Buffer init */
#define OFFSET 5
    InitBuffer(&b, data + OFFSET, 5, 3);

    i = CheckData(data, LENGTH_OF(data), 0, 0);

    PrintResult("InitCheck", i);
    PrintResult("Width", GetBufferWidth(&b) == 5);
    PrintResult("Used", GetBufferUsedColumns(&b) == 0);

    /* Fill 2..7 columns. */
    for (j = 2; j < 8; j++) {
        int c = (j <= 5) ? j : 5;
        printf("\nFill %d\n", j);
        SetData(values, LENGTH_OF(values), 20+j, 0);
        u = BufferBlockFill(&b, values, 3, j);
        PrintResult("Uncopied", u == j-c);
        PrintResult("Start", b.start == 0);
        PrintResult("Used", GetBufferUsedColumns(&b) == c);
        i = CheckData(data, OFFSET, 0, 0); /* First part untouched. */
        PrintResult("Data sanity1", i);
        i = CheckData(data + OFFSET, c*3, 20+j, 0);
        PrintResult("Data sanity2", i);
        i = CheckData(data + OFFSET + c*3, LENGTH_OF(data)-OFFSET-c*3, 0, 0);
        PrintResult("Data sanity3", i);
    }
}

static void TestBufferEmpty(void)
{
    double data[50];
    double values[50];
    double *dp;
    Buffer b;
    int s, u, dc, nc, i, col;

#define OFFSET 5
    SetData(data, LENGTH_OF(data), 0, 0);
    SetData(data + OFFSET, 3*5, OFFSET, 1); /* 5..20 */

    InitBuffer(&b, data + OFFSET, 5, 3); /* copy 5..20 into the buffer. */

    for (s = 0; s < 5; s++) { /* Start column in the buffer. */
        for (u = 1; u <= 5; u++) { /* Number of used columns. */

            for (dc = 0; dc < 6; dc++) { /* desired number of columns to copy. */
                printf("\nEmpty s%d, u%d, dc%d\n", s, u, dc);
                int ec = u; /* expected number of copied columns. */
                if (ec > dc) ec = dc;

                /* Copying data out from the buffer does not change the buffer, except for the
                 * b.used and b.start variables.
                 */
                b.start = s;
                b.used = u;

                SetData(values, LENGTH_OF(values), 100, 0);
                nc = BufferBlockEmpty(values + 5, 3, dc, &b); /* nc = not copied. */

                PrintResult("Not copied", nc == u-ec);

                i = CheckData(values, 5, 100, 0);
                i &= CheckData(values + 5 + ec*3, LENGTH_OF(values) - 5 - ec * 3, 100, 0);
                PrintResult("Out-ofbound sanity", i);

                col = s;
                i = TRUE;
                dp = values + 5;
                while (ec > 0) {
                    i &= CheckData(dp, 3, 5 + col*3, 1);
                    col++;
                    if (col == 5) col = 0;
                    dp += 3;
                    ec--;
                }
                PrintResult("Copied sanity", i);
            }
        }
    }
}

int main(void)
{
    TestBufferFill();
    printf("------------------------------------\n");
    TestBufferEmpty();
    return 0;
}
#endif
