
from acqgenlib import codebox, build_ast, soem_codegen, generic_codegen, function_blocks

class CodeData:
    """
    Helper class for generating code of acq.

    @ivar signal_num: First available number for creating unique signal variable names.
    @ivar signal_var_code: Code with signal variable declarations.
    @ivar signal_variables: Mapping of signals to code variables.

    @ivar block_num: First available number for creating unique block variable names.
    @ivar block_var_code: Code with block variable declarations.
    @ivar block_init_var_code: Code for initializing the block variables.
    @ivar block_init_code: Code for initializing block variables.
    @ivar block_exec_code: Code for performing an IO cycle in real-time context.

    @ivar soem_data: L{SoemCodeData}
    """
    def __init__(self):
        self.signal_num = 100 - 1
        self.signal_var_code = codebox.CodeBox()
        self.signal_variables = {}

        self.block_num = 500 - 1
        self.block_var_code = codebox.CodeBox()
        self.block_init_var_code = codebox.CodeBox()
        self.block_init_code = codebox.CodeBox()
        self.block_exec_code = codebox.CodeBox()

        self.soem_data = soem_codegen.SoemCodeData()

    def get_signal_num(self):
        self.signal_num = self.signal_num + 1
        return self.signal_num

    def add_signal_variable(self, signal, sigvar):
        assert signal not in self.signal_variables
        self.signal_variables[signal] = sigvar

    def get_signal_variable(self, signal):
        return self.signal_variables[signal]

    def get_block_num(self):
        self.block_num = self.block_num + 1
        return self.block_num

def create_signal_variables(signals, code_data):
    """
    Create 'double' variables for the data values of the given signals.

    @param signals: Signals to create variables for.
    @param code_data: Code generator helper class.
    """
    for sig in signals:
        varnum = code_data.get_signal_num()
        sigvar = "signal{:03d}".format(varnum)
        code_data.add_signal_variable(sig, sigvar)

        code_data.signal_var_code.append("double {}; /* Signal '{}' */".format(sigvar, sig))

def create_soem_inputs(block, block_prefix, code_data):
    """
    Create code for performing soem input (data read from soem).

    @param block: The soem input block to convert.
    @type  block: L{SoemInputs}

    @param code_data: Code generator helper class.
    """
    for signal in block.out_values:
        soem_val = code_data.soem_data.get_soemdata_pre(signal, code_data.block_exec_code)
        assert soem_val is not None
        sigvar = code_data.get_signal_variable(signal)
        if signal.iop_num is None:
            value = soem_val
        else:
            # Casting to uint64 won't work for some high bits, depending on used C compiler.
            assert signal.iop_num < 32 # Let's play safe.
            value = "(((uint64_t)({})) & (1LL << {})) != 0 ? 1.0 : 0.0"
            value = value.format(soem_val, signal.iop_num)

        code_data.block_exec_code.append("{} = {};".format(sigvar, value))

def create_soem_outputs(block,block_prefix, code_data):
    """
    Create code for performing soem output (data pushed to soem).

    @param block: The soem output block to convert.
    @type  block: L{SoemOutputs}

    @param code_data: Code generator helper class.
    """
    for signal in block.in_values:
        sigvar = code_data.get_signal_variable(signal)
        soem_val = code_data.soem_data.get_soemdata_pre(signal, code_data.block_exec_code)
        if soem_val is None:
            code_data.soem_data.soemdata_post_write(signal, sigvar, code_data.block_exec_code)
        else:
            if signal.iop_num is None:
                code_data.block_exec_code.append("{} = {};".format(soem_val, sigvar))
            else:
                line = "{} = (double)((uint64_t)({}) | ({} >= 0.5 ? (1LL << {}) : 0LL));"
                line = line.format(soem_val, soem_val, sigvar, signal.iop_num)
                code_data.block_exec_code.append(line)

            code_data.soem_data.soemdata_post_write(signal, soem_val, code_data.block_exec_code)


def create_output_code(block, block_prefix, code_data, config):
    """
    Create code for performing output from matlab.

    @param block: The 'from_matlab' block to convert.
    @type  block: L{FromMatlabBlock}

    @param code_data: Code generator helper class.
    @param config: Configuration of the data acquisition.
    """
    if config.buffered:
        ring_index = block.ring_index

        blk_index = code_data.get_block_num()
        block_var = "bufread{}".format(blk_index)
        block_obj = block_prefix + block_var
        col_var = "column{}".format(blk_index)
        col_obj = block_prefix + col_var

        # Buffer read structure and rescue column.
        code_data.block_var_code.add("BufferRead {};".format(block_var))
        code_data.block_init_code.add("InitBufferRead(&{}, {});".format(block_obj, len(block.out_values)))
        line = "double {}[{}]; /* Copy of last column. */".format(col_var, len(block.out_values))
        code_data.block_var_code.add(line)

        bxe = code_data.block_exec_code
        bxe.add("error = NULL;")
        bxe.add("rescue = FALSE;");
        line = "column = GetReadColumn(&{}, RING{}_PTR, &error);"
        bxe.add(line.format(block_obj, ring_index));
        bxe.add("if (column != NULL) {", post=1)
        bxe.add("if (error != NULL) {", post=1)
        line = "AddCollectError(&collected_errors, error, ET_WARNING, {}, sample_index);"
        bxe.add(line.format(ring_index))
        bxe.add("}", pre=-1)
        line = "rescue = !BufferReadHasUnread(&{}); /* Have column but no more unread -> last column! */"
        bxe.add(line.format(block_obj))
        bxe.add("} else {", pre=-1, post=1)
        bxe.add("if (error != NULL) {", post=1)
        line = "AddCollectError(&collected_errors, error, ET_ERROR, {}, sample_index);"
        bxe.add(line.format(ring_index))
        bxe.add("}", pre=-1)
        bxe.add("column = {}; /* No column, use rescue column instead. */".format(col_obj))
        bxe.add("}", pre=-1)

        bxe.append("if (rescue) {", post=1)
        for sig_idx, signal in enumerate(block.out_values):
            sigvar = code_data.get_signal_variable(signal)
            bxe.append('{}[{}] = {} = column[{}];'.format(col_obj, sig_idx, sigvar, sig_idx))
        bxe.add("} else {", pre=-1, post=1)
        for sig_idx, signal in enumerate(block.out_values):
            sigvar = code_data.get_signal_variable(signal)
            bxe.append('{} = column[{}];'.format(sigvar, sig_idx))
        bxe.add("}", pre=-1)

    else:
        # Unbuffered IO, (matlab) 'inputs' always exists.
        for sig_idx, signal in enumerate(block.out_values):
            sigvar = code_data.get_signal_variable(signal)
            code_data.block_exec_code.append('{} = inputs[{}];'.format(sigvar, sig_idx))

def create_input_code(block, block_prefix, code_data, config):
    """
    Create code for performing an input to matlab.

    @param block: The 'to_matlab' block to convert.
    @type  block: L{ToMatlabBlock}

    @param code_data: Code generator helper class.
    @param config: Configuration of the data acquisition.
    """
    if config.buffered:
        ring_index = block.ring_index

        blk_index = code_data.get_block_num()
        block_var = "bufwrite{}".format(blk_index)
        block_obj = block_prefix + block_var

        prewrite = max(0, block.prewrite)
        if prewrite < 1:
            # It's a fraction (or 0, but that's harmless)
            prewrite = int(prewrite * block.buffersize)

        prewrite = min(prewrite, block.buffersize - 1)
        code_data.block_var_code.append("BufferWrite {};".format(block_var))
        line = 'InitBufferWrite(&{}, {}, {});'
        code_data.block_init_code.append(line.format(block_obj, len(block.in_values), prewrite))

        if block.enable is None:
            condition = "FALSE"
        else:
            sigvar = code_data.get_signal_variable(block.enable)
            condition = "{} < 0.5".format(sigvar)

        bxe = code_data.block_exec_code
        bxe.add("error = NULL;")
        line = 'column = GetWriteColumn(&{}, RING{}_PTR, {}, &error);'
        bxe.add(line.format(block_obj, ring_index, condition))
        bxe.add("if (column != NULL) {", post=1)

        bxe.add("if (error != NULL) {", post=1)
        line = "AddCollectError(&collected_errors, error, ET_WARNING, {}, sample_index);"
        bxe.add(line.format(ring_index))
        bxe.add("}", pre=-1)

        for sig_idx, signal in enumerate(block.in_values):
            sigvar = code_data.get_signal_variable(signal)
            bxe.add('column[{}] = {};'.format(sig_idx, sigvar))

        bxe.add("} else if (error != NULL) {", pre=-1, post=1)
        line = "AddCollectError(&collected_errors, error, ET_ERROR, {}, sample_index);"
        bxe.add(line.format(ring_index))
        bxe.add("}", pre=-1)

    else:
        # Unbuffered IO, (matlab) 'outputs' always exists.
        for sig_idx, signal in enumerate(block.in_values):
            sigvar = code_data.get_signal_variable(signal)
            code_data.block_exec_code.append('outputs[{}] = {};'.format(sig_idx, sigvar))

def create_trigger_code(block, block_prefix, code_data):
    """
    Create code for performing a trigger.

    @param block: The trigger block to convert.
    @type  block: L{TriggerBlock}

    @param code_data: Code generator helper class.
    """
    blk_index = code_data.get_block_num()
    block_var = "trigger{}".format(blk_index)
    block_obj = block_prefix + block_var
    code_data.block_var_code.append("TriggerBlock {};".format(block_var))

    code_data.block_init_code.append('TriggerBlockInit(&{});'.format(block_obj))

    if block.policy in ('up', 'value'):
        up_trigger = "TRUE"
    else:
        up_trigger = "FALSE"

    if block.policy in ('down', 'value'):
        down_trigger = "TRUE"
    else:
        down_trigger = "FALSE"

    line = 'TriggerBlockSetPoint(&{}, {}, {}, {});'.format(block_obj, block.value, up_trigger, down_trigger)
    code_data.block_init_code.append(line)
    if block.interval is not None:
        line = 'TriggerBlockSetCount(&{}, {} / SOEM_INTERVAL);'.format(block_obj, block.interval)
        code_data.block_init_code.append(line)
    elif block.count is not None:
        line = 'TriggerBlockSetCount(&{}, {});'.format(block_obj, block.count)
        code_data.block_init_code.append(line)

    in_var = code_data.get_signal_variable(block.in_value)
    out_var = code_data.get_signal_variable(block.out_value)
    code_data.block_exec_code.append('TriggerBlockAddSample(&{}, {});'.format(block_obj, in_var))
    code_data.block_exec_code.append('{} = TriggerBlockEnabled(&{}) ? 1.0 : 0.0;'.format(out_var, block_obj))

def create_filter_code(block, block_prefix, code_data):
    """
    Create code for performing a filter.

    @param block: The filter block to convert.
    @type  block: L{FilterBlock}

    @param code_data: Code generator helper class.
    """
    blk_index = code_data.get_block_num()
    block_var = "filter{}".format(blk_index)
    block_obj = block_prefix + block_var
    code_data.block_var_code.append("FilterBlock {};".format(block_var))

    # Init code
    weights_var = "weights{}".format(blk_index)
    weights_vals = ", ".join(str(w) for w in block.weights)
    line = "static const double {}[] = {{{}}};".format(weights_var, weights_vals)
    code_data.block_init_var_code.append(line)
    code_data.block_init_code.append('FilterBlockInit(&{});'.format(block_obj))

    line = 'FilterBlockSetWeights(&{}, &mem_mgr, {}, {});'
    code_data.block_init_code.append(line.format(block_obj, weights_var, len(block.weights)))

    # Exec code
    in_var = code_data.get_signal_variable(block.in_value)
    out_var = code_data.get_signal_variable(block.out_value)
    line = "{} = FilterBlockAddSampleCompute(&{}, {});".format(out_var, block_obj, in_var)
    code_data.block_exec_code.append(line)

def create_function_code(block, block_prefix, code_data, func_gen):
    """
    Create code for computing a function.

    @param block: The function block to convert.
    @type  block: L{BaseFunctionBlock}

    @param code_data: Code generator helper class.
    @param func_gen: Function generator.
    """
    blk_index = code_data.get_block_num()
    if isinstance(block, build_ast.SineFunctionBlock):
        vardef, varinit, value = func_gen.generate_sine_function(block_prefix, block, blk_index)
    elif isinstance(block, build_ast.BlockFunctionBlock):
        vardef, varinit, value = func_gen.generate_block_function(block_prefix, block, blk_index)
    elif isinstance(block, build_ast.TriangleFunctionBlock):
        vardef, varinit, value = func_gen.generate_triangle_function(block_prefix, block, blk_index)
    else:
        assert False, "Unexpected kind of function"

    code_data.block_var_code.add_box(vardef)
    code_data.block_init_code.add_box(varinit)

    sigvar = code_data.get_signal_variable(block.out_value)
    code_data.block_exec_code.append("{} = {};".format(sigvar, value))


def create_block_code(blocks, code_data, config):
    """
    Generate code to perform data io according to the blocks, using the signal variable names.

    @param blocks: Blocks to execute, in order of execution.
    @param code_data: Code generator helper class.
    @param config: Configuration of the data acquisition.
    """
    if config.buffered:
        # Column pointer for input and output with matlab.
        has_fromML_block = any(isinstance(blk, build_ast.FromMatlabBlock) for blk in blocks)
        has_toML_block = any(isinstance(blk, build_ast.ToMatlabBlock) for blk in blocks)
        if has_fromML_block or has_toML_block:
            code_data.signal_var_code.add("double *column;")
            code_data.signal_var_code.add("const char *error;")
            code_data.signal_var_code.add("BOOL rescue;")

    # Construct function generator, and generate initialization of the sine sample arrays.
    func_gen = function_blocks.get_function_generator(blocks, config)
    vardef, varinit = func_gen.generate_sine_samples()
    code_data.block_init_code.add_box(vardef) # static sample arrays.
    code_data.block_init_code.add_box(varinit)

    block_prefix = "blockvars->"
    for block in blocks:
        create_signal_variables(block.get_out_values(), code_data)

        code_data.block_init_var_code.ensure_empty()
        code_data.block_init_code.ensure_empty()
        code_data.block_exec_code.ensure_empty()

        # Write a single line expressing what the execution code does.
        out_text = ", ".join(str(sig) for sig in block.get_out_values())
        in_text  = ", ".join(str(sig) for sig in block.get_in_values())
        line = []
        if len(out_text) > 0:
            line.append(out_text)
            line.append(" <- ")
        line.append(block.get_name())
        if len(in_text) > 0:
            line.append(" <- ")
            line.append(in_text)
        code_data.block_exec_code.append('/* {} */'.format("".join(line)))


        if isinstance(block, build_ast.SoemInputs):
            create_soem_inputs(block, block_prefix, code_data)
            continue

        if isinstance(block, build_ast.SoemOutputs):
            create_soem_outputs(block, block_prefix, code_data)
            continue

        if isinstance(block, build_ast.FromMatlabBlock):
            create_output_code(block, block_prefix, code_data, config)
            continue

        if isinstance(block, build_ast.ToMatlabBlock):
            create_input_code(block, block_prefix, code_data, config)
            continue

        if isinstance(block, build_ast.TriggerBlock):
            create_trigger_code(block, block_prefix, code_data)
            continue

        if isinstance(block, build_ast.FilterBlock):
            create_filter_code(block, block_prefix, code_data)
            continue

        if isinstance(block, build_ast.BaseFunctionBlock):
            create_function_code(block, block_prefix, code_data, func_gen)
            continue

        assert False, "Unknown block {} encountered".format(type(block))

def create_config_code(config):
    """
    Generate code for the general configuration.
    """
    box = codebox.CodeBox()

    # Generate global constants for the configuration.
    box.append("const size_t MEMSIZE = {}; /* bytes. */".format(config.memsize))
    box.append("const uint64_t SOEM_INTERVAL = {}; /* nano-seconds. */".format(config.delay))

    box.append("const char *NICS[] = {", post=1)
    box.append("\"{}\",".format(config.eth_port))
    box.append("NULL,")
    box.append("};", pre=-1)
    return box

    return {'global_config' : [box]}


def create_codelines(blocks, signals, config):
    """
    Generate blocks of code for the specifation.

    @param blocks: Blocks of the input specification.
    @type  blocks: C{list} of blocks.

    @param signals: Signals of the input specification.
    @type  signals: C{list} of L{Name}

    @return: Generated blocks of code to be merged into a template, ordered by name.
    @rtype:  C{dict} of C{str} to C{list} of L{CodeBox}.
    """
    code_data = CodeData()
    soem_codegen.create_soem_data(blocks, code_data)
    soem_codegen.create_soem_instance_cache(code_data)

    create_block_code(blocks, code_data, config)

    codelines = {
            "block_variables" : [code_data.block_var_code],
            "block_init" : [code_data.block_init_var_code, code_data.block_init_code],
            "block_exec" : [code_data.signal_var_code, code_data.block_exec_code],
            'global_config' : [create_config_code(config)],
    }

    codelines = generic_codegen.merge_codelines(codelines, code_data.soem_data.codelines)
    return codelines
