#!/usr/bin/env python3

import getopt, sys, os
from acqgenlib import report, build_ast, codegen, codebox, generic_codegen, signalgroup_codegen
from acqgenlib import mex_codegen, templates, compile_driver

VERSION = "@VERSION@"

def print_help():
    print("")
    print("acqgen: Compile a data acquisition configuration file to a Matlab module.")
    print("")
    print("acqgen [options] configuration")
    print("with options:")
    print("")
    print("  -h, --help")
    print("          Print this help")
    print("")
    print("  -l, --lines")
    print("          Generate #line pragmas in the generated file to assist locating")
    print("          the origin of output.")
    print("")
    print("  -t TEMPLATE, --template=TEMPLATE")
    print("          Use the TEMPLATE file for generating C code")
    print("")
    print("  -o CODE, --output=CODE")
    print("          Write output to CODE file")
    print("")
    print("  --no-mex")
    print("          Do not compile the generated file")
    print("")
    print("  --verbose")
    print("          Display Mex command being performed")
    print("")
    print("  --q, --quiet")
    print("          Do not print info lines")
    print("")
    if not VERSION.startswith("@"):
        print("Version: {}".format(VERSION))
        print("")

def run():
    short_opts = 'hqo:t:l'
    long_opts = ['help', 'lines', 'output=', 'template=', 'no-mex', 'verbose', 'quiet']
    try:
        opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
    except getopt.GetoptError as ex:
        report.report_error(None, str(ex))
        # Program dies.

    code_name = None
    template = None
    pragmas = False
    no_mex = False
    verbose = False
    quiet = False
    for opt, val in opts:
        if opt in ('-h', '--help'):
            print_help()
            sys.exit(0)

        if opt in ('-o', '--output'):
            code_name = val
            continue

        if opt in ('-l', '--lines'):
            pragmas = True
            continue

        if opt == '--verbose':
            verbose = True
            continue

        if opt in ('--quiet', '-q'):
            quiet = True
            continue

        if opt in ('-t', '--template'):
            if not os.path.isfile(val):
                report.report_error(None, "Template file \"{}\" is not an existing file".format(val))
                # Program dies.
            template = val
            continue

        if opt == '--no-mex':
            no_mex = True
            continue

        assert False, "Unknown option '{}' encountered".format(opt)

    report.SKIP_INFO = quiet


    if len(args) == 0:
        report.report_error(None, "Missing configuration file (use '-h' to get help)")
        # Program dies.
    elif len(args) > 1:
        report.report_error(None, "Found too many configuration files, exepcted 1, found {}".format(len(args)))
        # Program dies.

    spec_name = args[0]

    c_path = generic_codegen.derive_code_filename(code_name, spec_name, ".c")
    c_filename = os.path.basename(c_path)

    other_m = os.path.basename(generic_codegen.derive_code_filename(None, c_path, ".m"))
    other_mexes = os.path.basename(generic_codegen.derive_code_filename(None, c_path, ".mex"))
    mexa64_filename = other_mexes + "a64"

    output_dirname = os.path.dirname(other_mexes)
    if output_dirname == '':
        output_dirname = os.curdir


    # Remove output files if they exist, and warn about other .mex* files.
    for fname in os.listdir(output_dirname):
        if fname in (c_filename, mexa64_filename):
            fpath = os.path.join(output_dirname, fname)
            msg = "Removing output file '{}' before running Mex compiler".format(fpath)
            report.report_error(None, msg, type="info")
            os.unlink(fpath)
            continue

        if fname == other_m or fname.startswith(other_mexes):
            msg = "Warning: Found similarly named '{}' matlab file, make sure you run the right file!"
            report.report_error(None, msg.format(fname), type="warning")
            continue

    # Run the generator.
    blocks = build_ast.build_ast(spec_name)
    blocks, config = build_ast.get_config(blocks)
    build_ast.check_setup(blocks, config, config.buffered)

    edges = build_ast.build_graph(blocks)

    blocks, edges = build_ast.cleanup_blocks_signals(blocks, edges)
    ordered_blocks = build_ast.decide_block_computes(blocks, edges)

    # Generate code lines
    codelines = codegen.create_codelines(ordered_blocks, list(edges.keys()), config)

    groups = signalgroup_codegen.generate_signalgroups(ordered_blocks, config)
    group_codelines = signalgroup_codegen.signalgroups_gencode(groups, config)

    common_codelines = {'common_code' : [codebox.read_file(templates.get_common_code())]}
    mex_codelines = {'ml_entrypoint' : [mex_codegen.gen_entrypoint(ordered_blocks, config)]}

    codelines = generic_codegen.merge_codelines(codelines, group_codelines,
                                                common_codelines, mex_codelines)

    # Merge code lines, fill template
    template_name = template # Command-line
    if template_name is None:
        template_name = config.template # Configuration file
    if template_name is None: # Default template name.
        if config.buffered:
            template_name = templates.get_buffered_template()
        else:
            template_name = templates.get_unbuffered_template()

    code = generic_codegen.fill_template(codelines, template_name, pragmas)

    # Write to file.
    codebox.write_code(code, c_path)

    if not no_mex:
        # Compile the code
        ci = compile_driver.build_compile_info(config)
        ci.verbose = verbose
        compile_driver.mex_cc(c_path, ci)

if __name__ == '__main__':
    run()
