"""
Code generator for Mex interface functions.
"""
from acqgenlib import codebox, build_ast

def gen_header(code):
    code.add("/** Standard mex entry point. */")
    code.add("void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])")
    code.add("{", post=1)

def gen_footer(code):
    code.add('}', pre=-1)

def gen_die(code, id_name, message, suppress_unlock=False):
    """
    Generate code that throws a Matlab error and dies.
    """
    # Escape \ and " in the message.
    message = message.replace('\\', '\\\\').replace('"', '\\"')

    if not suppress_unlock:
        code.add("UnlockMex();")

    code.add('mexErrMsgIdAndTxt("{}",'.format(id_name), post=2)
    code.add('"{}");'.format(message), post=-2)

def gen_string_commands(code):
    """
    Recognize and process string command calls.
    """
    code.add("if (nlhs == 0 && nrhs == 1 && mxIsChar(prhs[0])) { /* Final report action. */", post=1)

    # Add "recognize Final report action" code.
    code.add('if (CompareCommand(mxGetChars(prhs[0]), mxGetN(prhs[0]), "report")) {', post=1)
    code.add('BOOL is_op = ml_state == MLS_OPERATIONAL;')
    code.add("UnlockMex();")
    code.add("/* At this point dynamic memory does not exist any more, and the function has been unlocked")
    code.add(" * and may disappear on exit.")
    code.add(" * Last chance to output informatoion!")
    code.add(" */")
    code.add("if (is_op) DumpCollectedErrors(&collected_errors);")
    code.add("return;")
    code.add("}", pre=-1)

    # Unrecognized string command, stop and give an error.
    gen_die(code, "acq:unknown:string-command",
            "Unrecognized string command found, use 'stop' to stop the data acquisition")

    code.add("}", pre=-1) #end of 'string command' recognition.
    code.add()

def gen_argument_checks(code, num_to, num_from, buffered):
    """
    Generate code for checking arguments.

    @param code: Code to append.
    @param num_to: Number of 'to_matlab' blocks in the configuration.
    @param num_from: Number of 'from_matlab' blocks in the configuration.
    @param buffered: Whether code for the buffered mode is being generated.
    """
    if buffered:
        matrix_desc = "cell matrix"
        gen_rhs_typecheck = lambda arg: "mxIsCell({})".format(arg)
    else:
        matrix_desc = "real-valued matrix"
        gen_rhs_typecheck = lambda arg: "mxIsDouble({}) && !mxIsComplex({})".format(arg, arg)

    # Check output (lhs side) count.
    if num_to > 0:
        code.add('/* Ensure output space exists. */')
        code.add('if (nlhs != 1) {', post=1)
        msg = "Incorrect number of output arguments, function has one output {} data argument"
        gen_die(code, "acq:incorrect:lhs:argument:count", msg.format(matrix_desc))
        code.add("}", pre=-1)
    else:
        code.add('/* Ensure output space does not exist. */')
        code.add('if (nlhs != 0) {', post=1)
        msg = "Incorrect number of output arguments, function does not expect any output argument"
        gen_die(code, "acq:incorrect:lhs:argument:count", msg)
        code.add("}", pre=-1)

    code.add()

    if num_from > 0:
        code.add('/* Ensure input exists. */')
        code.add('if (nrhs != 1) {', post=1)
        msg = "Incorrect number of input arguments, function requires one {} data argument"
        gen_die(code, "acq:incorrect:rhs:argument:count", msg.format(matrix_desc))
        code.add("}", pre=-1)

        # Check RHS type.
        code.add("if (!({})) {{".format(gen_rhs_typecheck("prhs[0]")), post=1)
        msg = "Incorrect input argument type, function requires 1 {} data argument"
        gen_die(code, "acq:incorrect:rhs:argument:type", msg.format(matrix_desc))
        code.add("}", pre=-1)
    else:
        code.add('/* Ensure no input is provided. */')
        code.add('if (nrhs != 0) {', post=1)
        msg = "Incorrect number of input arguments, function does not expect any input argument"
        gen_die(code, "acq:incorrect:rhs:argument:count", msg)
        code.add("}", pre=-1)

    code.add()

def gen_unbuffered_entrypoint(num_to, num_from):
    """
    Generate unbuffered standard Mex entry point function.

    @param num_to: Number of 'to_matlab' blocks.
    @type  num_to: C{int}

    @param num_from: Number of 'from_matlab' blocks.
    @type  num_to: C{from}
    """
    entry = codebox.CodeBox()
    gen_header(entry)
    gen_string_commands(entry)
    gen_argument_checks(entry, num_to, num_from, False)


    # Start data acquisition if it's not started yet.
    entry.add('if (ml_state == MLS_NOT_INITED) {', post=1)
    entry.add('InitErrorCollect(&collected_errors);')
    entry.add()
    entry.add('if (!LockMex()) {', post=1)
    # UnlockMex is not needed, since locking failed.
    entry.add('mexErrMsgIdAndTxt("acq:initialization:failed", "Failed to initialize the function");')
    entry.add('return;')
    entry.add('}', pre=-1)
    entry.add('/* ml_state has changed to MLS_MEM_LOCKED. */')
    entry.add()
    entry.add('InitDataCycleData();')
    entry.add('if (CollectHasError(&collected_errors)) {', post=1)
    entry.add('UnlockMex();')
    entry.add('/* ml_state has changed back to MLS_NOT_INITED. */')
    entry.add()
    entry.add('/* Init failed, no hope in continuing the process. */')
    entry.add('mexErrMsgIdAndTxt("acq:runtime", "%s", GetCollectErrorMessage(&collected_errors));')
    entry.add('return; /* never reached. */')
    entry.add('}', pre=-1)
    entry.add()
    entry.add("ml_state = MLS_OPERATIONAL; /* We're in business */")
    entry.add('}', pre=-1)
    entry.add()

    # Do the magic.
    params = []
    if num_to > 0:
        params.append("&plhs[0]")
    if num_from > 0:
        params.append("prhs[0]")

    entry.add("MatlabOperational({});".format(", ".join(params)))
    gen_footer(entry)
    return entry

def gen_buffered_entrypoint(num_to, num_from):
    """
    Generate buffered standard Mex entry point function.

    @param num_to: Number of 'to_matlab' blocks.
    @type  num_to: C{int}

    @param num_from: Number of 'from_matlab' blocks.
    @type  num_to: C{from}

    @return: Generated code.
    @rtype:  L{CodeBox}
    """
    entry = codebox.CodeBox()
    gen_header(entry)
    gen_string_commands(entry)
    gen_argument_checks(entry, num_to, num_from, True)


    # Start data acquisition if it's not started yet.
    entry.add('if (ml_state == MLS_NOT_INITED) {', post=1)
    entry.add('if (!LockMex()) {', post=1)
    # UnlockMex is not needed, since locking failed.
    entry.add('mexErrMsgIdAndTxt("acq:initialization:failed", "Failed to initialize the function");')
    entry.add('return;')
    entry.add('}', pre=-1)
    entry.add('ml_state = MLS_NOT_REALTIME;')
    entry.add()
    entry.add("InitErrorCollect(&collected_errors);")
    entry.add("ParentStartRealtime(&mesg_box, &RealTimeThread);")
    entry.add("ml_state = MLS_OPERATIONAL; /* We're in business */")
    entry.add('}', pre=-1)
    entry.add()

    # Do the magic.
    entry.add("if (!MatlabOperational(nlhs, plhs, nrhs, prhs)) {", post=1)
    entry.add("UnlockMex(); /* Real-time closed down, no point in staying up. */")
    entry.add("DumpCollectedErrors(&collected_errors);")
    entry.add('}', pre=-1)

    gen_footer(entry)

    return entry

def gen_entrypoint(blocks, config):
    """
    Generate the standard Mex main entry point.

    @param blocks: Blocks of the data acquisition.
    @param config: Configuration.

    @return: Generated code.
    @rtype:  L{CodeBox}
    """
    num_to = sum(1 for blk in blocks if isinstance(blk, build_ast.ToMatlabBlock))
    num_from = sum(1 for blk in blocks if isinstance(blk, build_ast.FromMatlabBlock))

    if config.buffered:
        return gen_buffered_entrypoint(num_to, num_from)
    else:
        return gen_unbuffered_entrypoint(num_to, num_from)
