import sys
from acqgenlib import build_ast, report, codebox

def check_unique_blockindex(blocks, klass, name):
    """
    Check that all block indices for blocks of type L{klass} are unique.

    @param blocks: Blocks of the input specification.
    @param klass: Class of blocks to check.
    @param name: Name of the block for the user.
    """
    used_indices = {}
    for blk in blocks:
        if not isinstance(blk, klass):
            continue

        if blk.index is None:
            msg = "Missing 'index' number for the '{}' block"
            report.report_error([blk.lineno], msg.format(name))


        double_defined = used_indices.get(blk.index)
        if double_defined is None:
            used_indices[blk.index] = blk
        else:
            blk_lines = [blk.lineno, double_defined.lineno]
            msg = "Index number {} used more than once in '{}' blocks"
            report.report_error(blk_lines, msg.format(blk.index, name))


def generate_signalgroups(blocks, config):
    """
    Collect and check the signal groups used in the blocks.

    @param blocks: Blocks of the input specification.
    @return: Signal groups ordered by index number.
    @rtype:  C{list} of (C{int}, L{ToMatlabBlock}/L{FromMatlabBlock})
    """
    # For the unbuffered case, don't bother the user with indices, simply
    # assign a number, as it is ignored anyway as there are no ring buffers.
    sorted_groups = []
    for block in blocks:
        if isinstance(block, build_ast.ToMatlabBlock) or isinstance(block, build_ast.FromMatlabBlock):
            sorted_groups.append((block.ring_index, block))

    if config.buffered:
        # Buffered case, verify unique block.index in input or output blocks.
        check_unique_blockindex(blocks, build_ast.ToMatlabBlock, 'to_matlab')
        check_unique_blockindex(blocks, build_ast.FromMatlabBlock, 'from_matlab')

    sorted_groups.sort()
    return sorted_groups


ML_INPUTGROUP = "MatlabInputGroup{}"
ML_OUTPUTGROUP = "MatlabOutputGroup{}"

# Buffered code generation
# {{{ def generate_ml_input(ringnum, grp):
def generate_ml_input(ringnum, grp):
    """
    Generate an 'InputGroupN' function that copies a matrix of signal group
    N to a buffer of that signal group.

    @return: Function code as a list of lines in a L{CodeBox}
    """
    rows = len(grp.in_values)
    funcname = ML_INPUTGROUP.format(ringnum)

    func_code = codebox.CodeBox()
    func_code.append("/**")
    func_code.append(" * Process signal group {}. Returns \\c true if data has been".format(ringnum))
    func_code.append(" * transferred or there is no valid data.")
    func_code.append(" */")
    func_code.append("static BOOL {}(mxArray *outcells)".format(funcname))
    func_code.append("{", post=1)
    func_code.append("Buffer *buffer;")
    func_code.append("mxArray *mat;")
    func_code.append()
    func_code.append("buffer = GetParentBuffer(RING{}_PTR);".format(ringnum))
    func_code.append("if (buffer == NULL) {", post=1)
    line = "if (UpdateParentRing(RING{}_PTR, 0) == 0) return FALSE; /* Really no buffer. */"
    func_code.append(line.format(ringnum))
    func_code.append("buffer = GetParentBuffer(RING{}_PTR);".format(ringnum))
    func_code.append("}", pre=-1)
    func_code.append()
    func_code.append("assert(buffer->used > 0);")
    func_code.append("mat = mxCreateDoubleMatrix({}, buffer->used, mxREAL);".format(rows))
    func_code.append("BufferBlockEmpty(mxGetPr(mat), {}, buffer->used, buffer);".format(rows))
    func_code.append()
    func_code.append("UpdateParentRing(RING{}_PTR, 1);".format(ringnum))
    func_code.append()
    func_code.append("mxSetCell(outcells, {}, mat);".format(grp.index - 1))
    func_code.append("return TRUE;")
    func_code.append("}", pre=-1)
    return func_code
# }}}
# {{{ def generate_ml_output(ringnum, grp):
def generate_ml_output(ringnum, grp):
    """
    Generate an 'OutputGroupN' function that copies a buffer of signal group N to a matrix.

    @return: Function code as a list of lines in a L{CodeBox}
    """
    rows = len(grp.out_values)
    funcname = ML_OUTPUTGROUP.format(ringnum)

    func_code = codebox.CodeBox()
    func_code.append("/**")
    func_code.append(" * Process signal group {}. Returns \\c true if data has been".format(ringnum))
    func_code.append(" * transferred or there is no valid data.")
    func_code.append(" */")
    func_code.append("static BOOL {}(const mxArray *in_mat)".format(funcname))
    func_code.append("{", post=1)
    func_code.append("Buffer *buffer;")
    func_code.append()
    line = "if (in_mat == NULL || !mxIsDouble(in_mat) || mxIsComplex(in_mat) || mxGetM(in_mat) != {}) return TRUE;"
    func_code.append(line.format(rows))
    func_code.append()
    func_code.append("buffer = GetParentBuffer(RING{}_PTR);".format(ringnum))
    func_code.append("if (buffer == NULL) {", post=1)
    line = "if (UpdateParentRing(RING{}_PTR, 0) == 0) return FALSE; /* Really no buffer. */"
    func_code.append(line.format(ringnum))
    func_code.append("buffer = GetParentBuffer(RING{}_PTR);".format(ringnum))
    func_code.append("}", pre=-1)
    func_code.append()
    func_code.append("BufferBlockFill(buffer, mxGetPr(in_mat), {}, mxGetN(in_mat));".format(rows))
    func_code.append("UpdateParentRing(RING{}_PTR, 1);".format(ringnum))
    func_code.append("return TRUE;")
    func_code.append("}", pre=-1)
    return func_code
# }}}
# {{{ def generate_ml_io_processing(groups):
def generate_ml_io_processing(groups):
    var_box = codebox.CodeBox()
    body_code = codebox.CodeBox()

    has_to, has_from = False, False
    largest_to_index = None
    for ringnum, grp in groups:
        if isinstance(grp, build_ast.FromMatlabBlock):
            has_from = True
        elif isinstance(grp, build_ast.ToMatlabBlock):
            has_to = True
            if largest_to_index is None or largest_to_index < grp.index:
                largest_to_index = grp.index

    if has_to or has_from:
        max_block = max(ringnum for ringnum, _ in groups)
        assert max_block < 64 # bitmask integer size limit.
        if max_block < 32:
            mask_type = "uint32_t"
        else:
            mask_type = "uint64_t"
    else:
        mask_type = None # No mask needed.
        max_block = 0

    # Compute mask constant for the output groups and synchronized input groups.
    allmask, syncmask = 0, 0
    for ringnum, grp in groups:
        allmask |= 1 << ringnum
        if isinstance(grp, build_ast.FromMatlabBlock):
            syncmask |= 1 << ringnum
        elif isinstance(grp, build_ast.ToMatlabBlock):
            if grp.synchronous:
                syncmask |= 1 << ringnum

    allmask_text = "0x{:x}".format(allmask)
    syncmask_text = "0x{:x}".format(syncmask)

    # Generate variables.
    if mask_type is not None:
        var_box.append("{} todo;".format(mask_type))

    if has_from:
        var_box.append("const mxArray *incells;")
    if has_to:
        var_box.append("mxArray *outcells;")

    # Generate body.

    if has_to:
        body_code.append("outcells = makeCellArray({});".format(largest_to_index))
        body_code.append("plhs[0] = outcells;")
    if has_from:
        body_code.append("incells = prhs[0];")

    # Generate the data copying loop for all groups.
    if has_from or has_to:
        # Construct text-fragments
        grps_texts = []
        for ringnum, grp in groups:
            bitval_text = "0x{:x}".format(1 << ringnum)
            todo_cond = "(todo & {}) != 0".format(bitval_text)
            if isinstance(grp, build_ast.ToMatlabBlock):
                funcname = ML_INPUTGROUP.format(ringnum)
                cells = "outcells"
                iocall_text = "{}({})".format(funcname, cells)
            else:
                funcname = ML_OUTPUTGROUP.format(ringnum)
                cells = "incells"
                iocall_text = "{}(mxGetCell({}, {}))".format(funcname, cells, grp.index - 1)

            texts = {'todo_cond': todo_cond,
                    'bitval_text': bitval_text,
                    'iocall_text': iocall_text,
                    }
            grps_texts.append(texts)

        # Generate looping output drop code.
        body_code.append()
        body_code.append("todo = {};".format(allmask_text))
        body_code.append("for (;;) { /* Loop until all output is dropped. */", post=1)
        for texts in grps_texts:
            line = "if ({todo_cond} && {iocall_text}) todo ^= {bitval_text};".format_map(texts)
            body_code.append(line)

        body_code.append()
        body_code.append("if ((todo & {}) == 0) break; /* All output and synchronized input done. */".format(syncmask_text))
        body_code.append()
        body_code.append("/* Wait for more space at the output, or for filled buffers at the input. */")
        body_code.append("if (ParentWaitMask(&mesg_box, (todo & {})) != PRS_NOTHING) break;".format(syncmask_text));
        body_code.append("}", pre=-1) # End 'for(;;)'
        body_code.append()

    # Block copying done, poll RT thread, and handle errors.
    body_code.append("return ParentPollBox(&mesg_box) != PRS_STOP_REALTIME;")

    return var_box, body_code
# }}}
# {{{ def generate_ml_input_output_funcs(groups):
def generate_ml_input_output_funcs(groups):
    """
    Generate input and output handling function for matlab.
    @return: code for the function foo(incell) -> outcell.
    """

    body_code = codebox.CodeBox()
    for ringnum, grp in groups:
        body_code.append()

        if isinstance(grp, build_ast.ToMatlabBlock):
            func_code = generate_ml_input(ringnum, grp)
        else:
            assert isinstance(grp, build_ast.FromMatlabBlock)
            func_code = generate_ml_output(ringnum, grp)

        body_code.add_box(func_code)

    return body_code
# }}}
# {{{ def generate_ringbuffer_code(groups):
def generate_ringbuffer_code(groups):
    num_groups = len(groups)

    ring_vars = codebox.CodeBox()
    ring_funcs = codebox.CodeBox()

    ring_vars.append("static Ringbuffer ringz[{}];".format(num_groups))

    ring_vars.append()
    ring_vars.append("/* Define RINGx macros to decouple access from the variable. */")
    for ringnum, _grp in groups:
        ring_vars.append("#define RING{} (ringz[{}])".format(ringnum, ringnum))
        ring_vars.append("#define RING{}_PTR (&ringz[{}])".format(ringnum, ringnum))

    ring_funcs.append("static BOOL InitRings(Memory *mem)")
    ring_funcs.append("{", post=1)

    if len(groups) > 0:
        ring_funcs.append("uint16_t i;")

        size_vars = []
        count_vars = []
        for ringnum, grp in groups:
            if isinstance(grp, build_ast.ToMatlabBlock):
                name = "input"
                values = grp.in_values
            else:
                name = "output"
                values = grp.out_values

            ring_funcs.append()
            ring_funcs.append("/* {} group {}: {} */".format(name, grp.index, ", ".join(str(v) for v in values)))
            ring_funcs.append("uint16_t rows{} = {};".format(ringnum, len(values)))
            ring_funcs.append("uint32_t cols{} = {};".format(ringnum, grp.buffersize))
            ring_funcs.append("size_t bufsize{} = rows{} * cols{};".format(ringnum, ringnum,ringnum))
            ring_funcs.append("uint16_t bufcount{} = {};".format(ringnum, grp.buffercount))
            count_vars.append("bufcount{}".format(ringnum))

            ring_funcs.append("size_t size{} = bufsize{} * bufcount{};".format(ringnum, ringnum, ringnum))
            size_vars.append('size{}'.format(ringnum))


        ring_funcs.append()
        ring_funcs.append("uint16_t sum_count = {};".format(" + ".join(count_vars)))
        ring_funcs.append("size_t sum_size = {};".format(" + ".join(size_vars)))
        ring_funcs.append("uint16_t buf_num = 0;")
        ring_funcs.append("Buffer *buffers = MemoryAllocate(mem, sizeof(Buffer) * sum_count);")
        ring_funcs.append("double *values = MemoryAllocate(mem, sizeof(double) * sum_size);")
        ring_funcs.append("if (buffers == NULL || values == NULL) return FALSE;")

        for ringnum, grp in groups:
            if isinstance(grp, build_ast.ToMatlabBlock):
                parent = "FALSE"
            else:
                parent = "TRUE"

            ring_funcs.append()
            ring_funcs.append("for (i = 0; i < bufcount{}; i++) {{".format(ringnum), post=1)
            line = "InitBuffer(&buffers[buf_num + i], values, cols{}, rows{});"
            ring_funcs.append(line.format(ringnum, ringnum))
            ring_funcs.append("values += bufsize{};".format(ringnum))
            ring_funcs.append("}", pre=-1)

            line = "InitRingbuffer(RING{}_PTR, &buffers[buf_num], bufcount{}, {});"
            ring_funcs.append(line.format(ringnum, ringnum, parent))
            ring_funcs.append("buf_num += bufcount{};".format(ringnum))

    # End of the 'InitRings' function.
    ring_funcs.append("return TRUE;")
    ring_funcs.append("}", pre=-1)
    ring_funcs.append()

    ring_funcs.append("static void DestroyRings(void)")
    ring_funcs.append("{", post=1)
    for ringnum, _grp in groups:
        ring_funcs.append("DestroyRingbuffer(RING{}_PTR);".format(ringnum))
    ring_funcs.append("}", pre=-1)
    ring_funcs.append()

    ring_funcs.append("/**")
    ring_funcs.append(" * Test whether one of the rings in the mask has buffers available in the")
    ring_funcs.append(" * parent stage.")
    ring_funcs.append(" * @note: This function may only be used if the tested ring buffers are")
    ring_funcs.append(" *      either not claimed at all, or claimed by the caller.")
    ring_funcs.append(" */")
    ring_funcs.append("BOOL RingsCheckMask(uint64_t mask)")
    ring_funcs.append("{", post=1)
    for ringnum, _ in groups:
        count_cond = "(RING{}.outbuffer_count > 0 || RING{}.parent_count > 0)".format(ringnum, ringnum)
        line = "if ((mask & (1 << {})) != 0 && {}) return TRUE;"
        ring_funcs.append(line.format(ringnum, count_cond))

    ring_funcs.append("return FALSE;")
    ring_funcs.append("}", pre=-1)

    return ring_vars, ring_funcs
# }}}

# Unbuffered code generation
def get_unbuf_ingrp_outgrp(groups):
    """
    Get the unbuffered input and the output group.

    @return: from and/or to signal groups, if they exist.
    @rtype:  C{tuple} (None or L{ToMatlabBlock}, None or L{FromMatlabBlock}).
    """
    to_groups  = [grp for _ringnum, grp in groups if isinstance(grp, build_ast.ToMatlabBlock)]
    from_groups = [grp for _ringnum, grp in groups if isinstance(grp, build_ast.FromMatlabBlock)]
    assert len(to_groups) + len(from_groups) == len(groups)

    if len(to_groups) > 1:
        linenos = [grp.lineno for grp in to_groups]
        msg = "Expected exactly one 'to_matlab' block, found {} blocks"
        report.report_error(linenos, msg.format(len(to_groups)))

    if len(from_groups) > 1:
        linenos = [grp.lineno for grp in from_groups]
        msg = "Expected exactly one 'from_matlab' block, found {} blocks"
        report.report_error(linenos, msg.format(len(from_groups)))

    if len(to_groups) == 0:
        to_group = None
    else:
        to_group = to_groups[0]

    if len(from_groups) == 0:
        from_group = None
    else:
        from_group = from_groups[0]

    return to_group, from_group

# {{{ def generate_unbuf_ml_operational(groups):
def generate_unbuf_ml_operational(groups):
    to_grp, from_grp = get_unbuf_ingrp_outgrp(groups)

    ml_op = codebox.CodeBox()
    params = []
    ml_op.add("/**")
    ml_op.add(" * Perform data IO with SOEM in unbuffered mode.")
    if to_grp is not None:
        ml_op.add(" * @param out_mat Result matrix to return to ML.")
        params.append("mxArray **out_mat")

    if from_grp is not None:
        ml_op.add(" * @param in_mat Mex argument with data. Unchecked.")
        params.append("const mxArray *in_mat")

    if len(params) == 0:
        params.append("void")

    ml_op.add(" */")
    ml_op.add("static void MatlabOperational({})".format(", ".join(params)))
    ml_op.add("{", post=1)

    if from_grp is not None:
        ml_op.add("const double *inputs;")
        inputs_var = "inputs"
    else:
        inputs_var = "NULL"

    if to_grp is not None:
        ml_op.add("double *outputs;")
        outputs_var = "outputs"
    else:
        outputs_var = "NULL"

    if from_grp is not None or to_grp is not None:
        ml_op.add()

    # Check and setup the 'inputs' value in C for reading the user data.
    if from_grp is not None:
        from_rows = len(from_grp.out_values)
        ml_op.add('/* Input is known to be a real-valued matrix, check it has enough rows. */')
        ml_op.add("if (mxGetM(in_mat) < {}) {{".format(from_rows), post=1)
        ml_op.add("UnlockMex();")
        ml_op.add('mexErrMsgIdAndTxt("acq:incorrect-size:rhs:matrix",')
        line = '                  "from_matlab data does not have enough rows");'
        ml_op.add(line)
        ml_op.add("}", pre=-1)

        ml_op.add("inputs = mxGetPr(in_mat);")
        ml_op.add()

    # Setup 'outputs' value in C for writing the results.
    if to_grp is not None:
        to_rows = len(to_grp.in_values)
        ml_op.add("/* Construct data output matrix. */")
        ml_op.add("*out_mat = mxCreateDoubleMatrix({}, 1, mxREAL);".format(to_rows))
        ml_op.add('outputs = mxGetPr(*out_mat);')

        ml_op.add()
    ml_op.add("PerformDataCycle({}, {});".format(inputs_var, outputs_var))
    ml_op.add("if (CollectHasFatalError(&collected_errors)) {", post=1)
    ml_op.add("UnlockMex();")
    ml_op.add('mexErrMsgIdAndTxt("acq:runtime", "%s", GetCollectErrorMessage(&collected_errors));')
    ml_op.add("}", pre=-1)

    ml_op.add("}", pre=-1) # End of function.

    return ml_op
# }}}

def signalgroups_gencode(groups, config):
    """
    Generate code for the signal groups.

    @param groups: Matlab input and output groups.
    @param config: Configuration.
    @return: Mapping of code-names to lists of code.
    """
    if config.buffered:
        ml_io_funcs = generate_ml_input_output_funcs(groups)
        ml_io_vars, ml_io_code = generate_ml_io_processing(groups)
        ring_vars, ring_funcs = generate_ringbuffer_code(groups)

        return {'ml_input_output_funcs': [ml_io_funcs],
                'ml_io_vars': [ml_io_vars],
                'ml_io_code': [ml_io_code],
                'ringbuffers_variable' : [ring_vars],
                'ringbuffer_funcs': [ring_funcs],
        }

    else:
        # Unbuffered code generation
        ml_op = generate_unbuf_ml_operational(groups)
        return {'ml_operational': [ml_op]}

