
from acqgenlib import codebox, build_ast, names

class SoemCodeData:
    """
    Helper class for generating the Soem interface code.

    @ivar io_num: First available number for creating unique io variable names.
    @ivar io_accesses: Mapping of hardware signals to L{SoemAccessPoint}
    @ivar soem_points: Mapping of io instances to L{SoemAccessPoint} (internal use only)
    @ivar codelines: Boxes with code, for each named method.
    """
    def __init__(self):
        self.io_num = 200 - 1
        self.io_accesses = {}
        self.soem_points = {}
        self.codelines = {}

    def get_io_num(self):
        self.io_num = self.io_num + 1
        return self.io_num

    def add_io_access(self, signal):
        """
        Register a hardware signal for generating soem access interface variables and code.
        """
        # Find soem access point through the signal
        soem_point = self.io_accesses.get(signal)
        if soem_point is not None:
            soem_point.signals.append(signal)
            return

        # Signal is unknown, check if the soem access point exists for other signals.
        # Create the normalized io instance.
        if signal.iop_num is None:
            ioinstance = signal
        else:
           ioinstance = names.Name(signal.lineno, signal.dev_name, signal.dev_num, signal.iop_name, None)

        # Find soem or create access point.
        soem_point = self.soem_points.get(ioinstance)
        if soem_point is None:
            soem_point = SoemAccessPoint(self.get_io_num(), ioinstance)
            self.soem_points[ioinstance] = soem_point

        self.io_accesses[signal] = soem_point
        soem_point.signals.append(signal)

    def get_soemdata_pre(self, signal, codebox):
        """
        For an already added signal, get the expression to get the curent soem value.
        """
        soem_point = self.io_accesses[signal]
        return soem_point.get_soemdata_pre(codebox)

    def soemdata_post_write(self, signal, value, codebox):
        """
        For an already added signal, provide an opportunity for soem to write the signal value.
        """
        soem_point = self.io_accesses[signal]
        soem_point.soemdata_post_write(value, codebox)

class SoemAccessPoint:
    """
    Data about accessing a Soem io instance.

    @ivar number: Unique number for creating variables.
    @type number: C{int}

    @ivar io_instance: Io instance represented by the object.
    @type io_instance: L{Name}

    @ivar signals: Signals accessing the instance.
    @type signals: C{list} of L{Name}

    @ivar access_type: 'input' or 'output'.
    @ivar access_type: C{str}

    @ivar num_accesses: Number of processed access from blocks so far (0 to len(signals))
    @type num_accesses: C{int}
    """
    def __init__(self, number, io_instance):
        self.number = number
        self.io_instance = io_instance
        self.signals = []
        self.num_accesses = 0

        if io_instance.get_inputkind() is not None:
            self.access_type = "input"
        else:
            assert io_instance.get_outputkind() is not None
            self.access_type = "output"

        self.io_instance_var = None
        self.cache_var = None

    def get_soem_read(self):
        """
        Internal routine to read the io instance from Soem.

        @return: Expression that reads the data from the Soem io instance.
        """
        assert self.access_type == 'input'
        return "SOEM_READ({})".format(self.io_instance_var)

    def get_soem_write(self, value, codebox):
        """
        Internal routine to generate code for writing the given expression value to Soem.
        """
        assert self.access_type == 'output'
        codebox.append("SOEM_WRITE({}, {});".format(self.io_instance_var, value))

    def get_soemdata_pre(self, codebox):
        """
        Code wants to use the io instance of Soem. Prepare for doing so.

        @return: If appropriate, return a data value to use for read or write.
                C{None} means the code is free to create its own value.
        """
        assert self.num_accesses < len(self.signals)
        self.num_accesses = self.num_accesses + 1

        if self.access_type == 'input':
            if self.num_accesses == 1:
                # For input, code needs to read the value the first time.
                if self.cache_var is not None:
                    codebox.append("{} = {};".format(self.cache_var, self.get_soem_read()))
                    return self.cache_var
                else:
                    return self.get_soem_read()
            else:
                # For next calls, return the cache.
                assert self.cache_var is not None
                return self.cache_var

        else:
            assert self.access_type == 'output'
            if self.cache_var is not None:
                # For output, cache may need to be cleared before first use.
                if self.num_accesses == 1:
                    codebox.append("{} = 0.0;".format(self.cache_var))

                return self.cache_var
            else:
                # Let code construct its own value
                return None

    def soemdata_post_write(self, value, codebox):
        """
        Code has constructed a value that needs to be written to Soem.

        @note: Code assumes that a cache variable given out from
                L{get_soemdata_pre} is returned here.
        """
        assert self.access_type == 'output'

        if self.num_accesses < len(self.signals):
            return  # Wait until after the last access.

        self.get_soem_write(value, codebox)


def create_soem_data(blocks, code_data):
    """
    Construct mapping of accessing soem as preparation for code generation.

    @param blocks Blocks of the specification.
    @return Soem helper data for code generation of soem access.
    """
    # Collect hardware signals from blocks.
    signals = []
    for block in blocks:
        if isinstance(block, build_ast.SoemInputs):
            signals.extend(block.out_values)

        elif isinstance(block, build_ast.SoemOutputs):
            signals.extend(block.in_values)

    soem_data = code_data.soem_data
    for sig in signals:
        soem_data.add_io_access(sig)

    # Generate body of finding the user io points from SOEM discovered slaves / IO points.
    # Also generate io_instanceXYZ variable declarations and clearing code for the former.
    io_variable_lines = []
    io_clear_variable_lines = []
    io_var_init_code = codebox.CodeBox()
    object_prefix = "instances->"
    for sap in sorted(soem_data.soem_points.values(), key=lambda x: x.io_instance):
        sig = sap.io_instance

        # Generate IoInstance variables pointing at the found entry.
        io_var = "io_instance{}".format(sap.number)
        sap.io_instance_var = object_prefix + io_var

        io_variable_lines.append("IoInstance *{}; /* Signal '{}' */".format(io_var, sig))
        io_clear_variable_lines.append("{}{} = NULL;".format(object_prefix, io_var))

        # Generate initialization code for the io instance variables.
        if sig.dev_name is not None:
            line = "{}{} = SoemFindIoInstance(ssd, \"{}\", {}, \"{}\");"
            line = line.format(object_prefix, io_var, sig.dev_name, sig.dev_num, sig.iop_name)
        else:
            kind = sig.get_inputkind()
            if kind is None:
                kind = sig.get_outputkind()
            assert kind is not None
            kind = kind.upper()
            line = "{}{} = SoemFindIoKind(ssd, DK_{}, {});".format(object_prefix, io_var, kind, sig.dev_num)

        io_var_init_code.add(line)
        io_var_init_code.add("if ({}{} == NULL) {{".format(object_prefix, io_var), post=1)
        line = "return \"Failed to find io point '{}', aborting\\n\";".format(sig)
        io_var_init_code.add(line)
        io_var_init_code.add("}", pre=-1)
        io_var_init_code.add()

    io_var_init_code.add("return NULL;")

    # Add generated code to the code lines.
    box = codebox.CodeBox()
    io_variable_lines.sort()
    box.extend(io_variable_lines)
    soem_data.codelines['io_variables'] = [box]

    box = codebox.CodeBox()
    io_clear_variable_lines.sort()
    box.extend(io_clear_variable_lines)
    soem_data.codelines['clear_io_vars'] = [box]

    soem_data.codelines['io_var_init'] = [io_var_init_code]

    return soem_data

def create_soem_instance_cache(code_data):
    """
    Create a cache variable for Soem access points that are used more than once.
    """
    for sap in sorted(code_data.soem_data.soem_points.values(), key=lambda x: x.io_instance):
        # Needs a cache when 2 signals exist in the specification, or when it is partly written.
        if len(sap.signals) > 1 or sap.signals[0].is_partly_write():
            cache_var = "soem_cache{}".format(sap.number)
            sap.cache_var = cache_var

            line = 'double {}; /* cache for {} */'.format(cache_var, sap.io_instance)
            code_data.signal_var_code.append(line)
