/**
 * Compare the given command with an expected value, and return whether they match case-insensively.
 * @param cmd Entered command.
 * @param length Number of characters in \a cmd.
 * @param value Expected value to compare against.
 * @return Expected value was detected (modulo case).
 */
static BOOL CompareCommand(const mxChar *cmd, size_t length, const char *value)
{
    while (length > 0 && *value != '\0') {
        char k = *cmd;
        if (k >= 'A' && k <= 'Z') k += 'a' - 'A';
        if (k != *value) return FALSE;

        cmd++;
        length--;
        value++;
    }
    return *value == '\0';
}

/** Names of the various error types, 'Unknown' and 'Huh' should never happen. */
static const char *type_descs[] = {
    "Unknown", "Warning", "Error", "Fatal", "Huh"
};

/**
 * Output the errors stored in the error collection to the console.
 * @param collect Error collection to print.
 */
static void DumpCollectedErrors(const ErrorCollect *collect)
{
    BOOL printed_errors = FALSE;
    int err_type = ET_FATAL;
    for (;;) {
        /* Print errors of type 'err_type'. */
        size_t i;
        for (i = 0; i < collect->num_recorded; i++) {
            const ErrorInfo *info = &collect->occurred[i];
            if (info->err_type != err_type) continue;

            printed_errors = TRUE;
            if (info->group_index >= 0) {
                mexPrintf("%s at stream %d: '%s'", type_descs[err_type], info->group_index, info->err_text);
            } else {
                mexPrintf("%s: '%s'", type_descs[err_type], info->err_text);
            }

            mexPrintf(" occurred %d time%s", info->num_occurrences,
                      (info->num_occurrences == 1 ? "" : "s"));

            if (info->first_index >= 0) {
                if (info->last_index >= 0) {
                    mexPrintf(" at first sample index %d and last sample index %d",
                              info->first_index, info->last_index);
                } else {
                    mexPrintf(" at sample index %d", info->first_index);
                }
            }
            mexPrintf("\n");
        }

        /* Continue with printing next type of errors. */
        if (err_type == ET_NONE) break;
        err_type--;
    }

    /* Make explicit that no messages to report exist. */
    if (!printed_errors) {
        mexPrintf("No messages to report at all!\n");
    }
}

/**
 * Copy a Soem error to #collected_errors.
 * @param ssd Soem state data.
 * @param collect Error collection to copy to.
 * @param sample_index Index number of the sample.
 */
static void CopySoemError(SoemStateData *ssd, ErrorCollect *collect, int32_t sample_index)
{
    const char *err_msg;
    if (GetSoemStateError(ssd) != NULL) {
        err_msg = GetSoemStateError(ssd);
    } else {
        /* A SOEM function failed, but no error was reported. */
        err_msg = "Unknown SOEM error";
    }
    AddCollectError(collect, err_msg, ET_FATAL, -1, sample_index);
}

/**
 * Initialize the data cycle data, returning whether the initialization worked, while
 * dumping any error into the #collected_errors.
 */
static void InitDataCycleData(void)
{
    const NanoTime INITIAL_DELAY = MakeNanos(0, 1000); /**< Initial delay 1ms. */

    const char *err_msg;

    pthread_mutex_init(&wait_lock, PTHREAD_MUTEX_NORMAL); /* Don't check on recursive locking. */
    pthread_mutex_lock(&wait_lock); /* Make the mutex unavailable. */

    sample_index = -1; /* Pre sampling stage. */

    /* Setup soem. */
    InitializeSoemState(&ssd);

    if (!SoemEngineWakeup(&ssd, &mem_mgr, NICS)) {
        CopySoemError(&ssd, &collected_errors, -1);
        DestroyDataCycleData();
        return;
    }

    if (!SoemSetupIoPoints(&ssd, &mem_mgr)) {
        CopySoemError(&ssd, &collected_errors, -1);
        DestroyDataCycleData();
        return;
    }

    err_msg = InitializeIoInstances(&ssd, &io_instances);
    if (err_msg != NULL) {
        AddCollectError(&collected_errors, err_msg, ET_FATAL, -1, -1);
        DestroyDataCycleData();
        return;
    }

    if (!InitializeBlockVariables(&block_vars)) {
        err_msg = "Failed to initialize block variables";
        AddCollectError(&collected_errors, err_msg, ET_FATAL, -1, -1);
        DestroyDataCycleData();
        return;
    }

    if (!SoemEngineStart(&ssd)) {
        CopySoemError(&ssd, &collected_errors, -1);
        DestroyDataCycleData();
        return;
    }

    /* Setup time until the first data exchange. */
    sample_time = NowNanos();
    IncreaseNano(&sample_time, &INITIAL_DELAY);
    increment = MakeNano64(SOEM_INTERVAL);
    sample_index = 0;
}
