#!/usr/bin/env python3
# Install script for acqgen
import sys, getopt, os, shutil

if sys.version_info.major < 3 or sys.version_info.minor < 5:
    print("ERROR: Python version 3.5 or newer required.")
    sys.exit(1)

version = "release-V1"
assert not version.startswith("@")

if not os.path.isfile("bin/acqgen"):
    print("ERROR: Missing 'bin/acqgen', running from wrong directory?")
    sys.exit(1)

long_opts = ["install-prefix=", "bin-dir=", "application-dir=", "soem-base=", "help",
             "i-know-what-i-am-doing"]
short_opts = "h"
try:
    opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
except getopt.GetoptError as exc:
    print("ERROR: {}".format(exc))
    sys.exit(1)

install_prefix = None
bin_dir = None
application_dir = None
soem_base = None
force = False
for opt, optval in opts:
    if opt in ('-h', '--help'):
        print("install 'acqgen' version {}.".format(version))
        print("Options:")
        print(" -h, --help             Display this online help")
        print(" --install-prefix=DIR   Install in DIR/bin, and DIR/share/acqgen")
        print("                        Default: DIR=/usr/local")
        print(" --bin-dir=BIN          Install binaries in BIN directory")
        print(" --application_dir=APP  Install application data in APP directory")
        print(" --soem-base=SOEM       Installed SOEM library to use")
        print(" --force                Force install even with failed SOEM library detection")
        sys.exit(0)

    if opt == "--install-prefix":
        install_prefix = os.path.abspath(optval)
        continue

    if opt == "--bin-dir":
        bin_dir = os.path.abspath(optval)
        continue

    if opt == "--application_dir":
        application_dir = os.path.abspath(optval)
        continue

    if opt == "--soem-base":
        soem_base = os.path.abspath(optval)
        continue

    if opt == " --force":
        force = True
        continue

    assert False, "Unknown option {} encountered".format((opt, optval))

if install_prefix is None:
    install_prefix = "/usr/local"

if bin_dir is None:
    bin_dir = os.path.join(install_prefix, "bin")

if application_dir is None:
    application_dir = os.path.join(install_prefix, "share/acqgen")

if soem_base is None:
    print("ERROR: Missing installed SOEM directory, set with --soem-base")
    sys.exit(1)

soem_lib = os.path.join(soem_base, "lib/libsoem.a")
if not os.path.isfile(soem_lib) and not force:
    print("ERROR: Installed SOEM path is incorrect, {} is expected to exist, but does not".format(soem_lib))
    sys.exit(1)

def modify_file(fname, repl_interpreter, patterns):
    handle = open(fname, "rt")
    lines = handle.readlines()
    handle.close()

    for idx in range(len(lines)):
        line = lines[idx]

        replaced = False
        if idx == 0 and repl_interpreter:
            if sys.executable is not None and len(sys.executable) > 0 and line.startswith("#!"):
                line = "#!{}\n".format(sys.executable)
                replaced = True

        for pat, repl in patterns.items():
            i = line.find(pat)
            if i < 0:
                continue

            line = line[:i] + repl + line[i+len(pat):]
            replaced = True

        if replaced:
            lines[idx] = line

    handle = open(fname, "wt")
    handle.writelines(lines)
    handle.close()

# Copy bin + modify bin/acqgen with supplied data
os.makedirs(bin_dir, exist_ok=True)
dest = os.path.join(bin_dir, "acqgen")
shutil.copy2("bin/acqgen", dest)
patterns = {"@" + "SHARED" + "@": application_dir}
modify_file(dest, True, patterns)

# Copy acqgenlib tree + set soem-base.
os.makedirs(application_dir, exist_ok=True)
shutil.copytree("acqgenlib", os.path.join(application_dir, "acqgenlib"))

dest = os.path.join(application_dir, "acqgenlib/code/__init__.py")
patterns = {"@" + "SOEM-INSTALL" + "@": soem_base}
modify_file(dest, False, patterns)

dest = os.path.join(application_dir, "acqgenlib/main.py")
patterns = {"@" + "VERSION" + "@": version}
modify_file(dest, False, patterns)
